#
#  This file is part of bioservices software
#
#  Copyright (c) 2013-2014 - EBI-EMBL
#
#  File author(s):
#      Thomas Cokelaer <cokelaer@ebi.ac.uk>
#
#
#  Distributed under the GPLv3 License.
#  See accompanying file LICENSE.txt or copy at
#      http://www.gnu.org/licenses/gpl-3.0.html
#
#  website: https://github.com/cokelaer/bioservices
#  documentation: http://packages.python.org/bioservices
#
##############################################################################
"""Interface to the EUtils web Service.

.. topic:: What is EUtils ?

    :URL: http://www.ncbi.nlm.nih.gov/books/NBK25499/
    :URL: http://www.ncbi.nlm.nih.gov/books/NBK25500/#chapter1.Demonstration_Programs

    .. highlights::

        The Entrez Programming Utilities (E-utilities) are a set of eight server-side programs that provide a stable interface into the Entrez query and database system at the National Center for Biotechnology Information (NCBI). The E-utilities use a fixed URL syntax that translates a standard set of input parameters into the values necessary for various NCBI software components to search for and retrieve the requested data. The E-utilities are therefore the structured interface to the Entrez system, which currently includes 38 databases covering a variety of biomedical data, including nucleotide and protein sequences, gene records, three-dimensional molecular structures, and the biomedical literature.

       -- from http://www.ncbi.nlm.nih.gov/books/NBK25497/, March 2013

"""
import json
from bioservices import REST
from bioservices import __version__
from bioservices import logger
logger.name = __name__


__all__ = ["EUtils", "EUtilsParser"]

# source:
# http://www.dalkescientific.com/writings/diary/archive/2005/09/30/using_eutils.html


class EUtils(REST):
    """Interface to `NCBI Entrez Utilities <http://www.ncbi.nlm.nih.gov/entrez>`_ service

    .. note:: Technical note: the WSDL interface was dropped in july 2015
        so we now use the REST service.

    .. warning:: Read the `guidelines
        <http://www.ncbi.nlm.nih.gov/books/NBK25497/>`_ before sending requests.
        No more than 3 requests per seconds otherwise your IP may be banned.
        You should provide your email by filling the :attr:`email` so that
        before being banned, you may be contacted.

    There are a few methods such as :meth:`ELink`, :meth:`EFetch`.
    Here is an example on how to use :meth:`EFetch` method to retrieve the
    FASTA sequence of a given identifier (34577063)::

        >>> from bioservices import EUtils
        >>> s = EUtils()
        >>> print(s.EFetch("protein", "34577063", rettype="fasta"))
        >gi|34577063|ref|NP_001117.2| adenylosuccinate synthetase isozyme 2 [Homo sapiens]
        MAFAETYPAASSLPNGDCGRPRARPGGNRVTVVLGAQWGDEGKGKVVDLLAQDADIVCRCQGGNNAGHTV
        VVDSVEYDFHLLPSGIINPNVTAFIGNGVVIHLPGLFEEAEKNVQKGKGLEGWEKRLIISDRAHIVFDFH
        QAADGIQEQQRQEQAGKNLGTTKKGIGPVYSSKAARSGLRMCDLVSDFDGFSERFKVLANQYKSIYPTLE
        IDIEGELQKLKGYMEKIKPMVRDGVYFLYEALHGPPKKILVEGANAALLDIDFGTYPFVTSSNCTVGGVC
        TGLGMPPQNVGEVYGVVKAYTTRVGIGAFPTEQDNEIGELLQTRGREFGVTTGRKRRCGWLDLVLLKYAH
        MINGFTALALTKLDILDMFTEIKVGVAYKLDGEIIPHIPANQEVLNKVEVQYKTLPGWNTDISNARAFKE
        LPVNAQNYVRFIEDELQIPVKWIGVGKSRESMIQLF

    Most of the methods take a database name as input. You can obtain the
    valid list by checking the :attr:`databases` attribute.

    A few functions takes Identifier(s) as input. It could be a list of
    strings, list of numbers, or a string where identifiers are separated
    either by comma or spaces.

    A few functions take an argument called **term**. You can use the **AND**
    keyword with spaces or + signs as separators::

        Correct:   term=biomol mrna[properties] AND mouse[organism]
        Correct:   term=biomol+mrna[properties]+AND+mouse[organism]

    Other special characters, such as quotation marks (") or the # symbol used
    in referring to a query key on the History server, could be represented by
    their URL encodings (%22 for "; %23 for #) or verbatim .::

        Correct: term=#2+AND+"gene in genomic"[properties]
        Correct: term=%232+AND+%22gene+in+genomic%22[properties]

    For information about retmode and retype, please see:

    http://www.ncbi.nlm.nih.gov/books/NBK25499/table/chapter4.T._valid_values_of__retmode_and/?report=objectonly


    """
    def __init__(self, verbose=False, email="unknown", cache=False,
                xmlparser="EUtilsParser"):
        url = "https://eutils.ncbi.nlm.nih.gov/entrez/eutils"
        super(EUtils, self).__init__(name="EUtils", verbose=verbose, url=url,
            cache=cache, requests_per_sec=3)

        warning = """

        NCBI recommends that users post no more than three URL requests per
        second. Failure to comply with this policy may result in an IP address
        being blocked from accessing NCBI. If NCBI blocks an IP address,
        service will not be restored unless the developers of the software
        accessing the E-utilities register values of the tool and email
        parameters with NCBI. The value of email will be used only to contact
        developers if NCBI observes requests that violate our policies, and we
        will attempt such contact prior to blocking access.  For more details
        see http://www.ncbi.nlm.nih.gov/books/NBK25497/#chapter2.chapter2_table1
        BioServices limits requests to 3 per seconds for this services.
        If you choose to set to a higher rate, this will be the user
        responsability. Within
        BioServices, we fill the parameter **tool** and **email**, however,
        to fill the latter you should provide your email either globablly
        when instanciating EUtils, or locally when calling a method.

        This message will not appear if you set the email as a parameter::

            e = EUtils(email="name@adress")

        or in you bioservices configuration file
        (.config/bioservices/bioservices.cfg) under linux with a user section::

            [user]
            email = yourname@somewhere


        """
        self._xmlparser = xmlparser

        self._databases = None
        self.tool = "BioServices, " + __version__

        #: fill this with your email address
        self.email = email
        if self.email == "unknown":
            # trying the bioservices config file
            if self.settings.params['user.email'][0] != "unknown":
                self.email = self.settings.params['user.email'][0]
            else:
                self.logging.warning(warning)

    def help(self):
        """Open EUtils help page"""
        self.on_web('http://www.ncbi.nlm.nih.gov/books/NBK25497')

    def _get_databases(self):
        """alias to run_eInfo"""
        # Let us use the REST services instead of WSDL, which fails sometimes
        # and for sure since version Sept 2015
        if self._databases is None:
            res = self.http_get('einfo.fcgi', params={'retmode':'json'})
            databases = res['einforesult']['dblist']

            self._databases = sorted(databases)
        return self._databases
    databases = property(_get_databases, doc="Returns list of valid databases")

    def _check_db(self, db=None):
        msg = "You must provide a valid databases from : "
        if db is None or db not in self.databases:
            raise ValueError(msg, self.databases)

    def _check_retmode(self, retmode, valids=['xml', 'text']):
        if retmode not in valids:
            raise ValueError("You must provide a retmode in %s" % valids)

    def _get_params(self, keys=[], **kargs):
        # could use a defaultdict from collections.
        params = {'tool': self.tool, 'email': self.email}
        # fill the structure with None
        for this in keys:
            params[this] = None
        # update structure with user's items if any

        for k, v in kargs.items():
            if k in keys:
                params[k] = v
            else:
                # unknown so let use it but raise a warning
                params[k] = v
                self.logging.warning("%s does not seem to be a known parameter. " % k+
                        "Use it anyway but may be ignored")
        return params

    def _get_einfo_params(self, **kargs):
        params = self._get_params(['db', 'version', 'retmode'], **kargs)
        return params

    def _get_esummary_params(self, **kargs):
        keys = ['WebEnv', 'query_key', 'retstart', 'retmax',
            'retmode', 'version']
        params = self._get_params(keys, **kargs)
        return params

    def _get_esearch_params(self, **kargs):
        keys = ['retmax', 'retstart', 'WebEnv', 'query_key',
                'datetype', 'retmode',
                'field', 'maxdate', 'mindate', 'reldate', 'rettype',
                'sort',  'usehistory']
        params = self._get_params(keys, **kargs)
        return params

    def _get_egquery_params(self, **kargs):
        params = self._get_params([], **kargs)
        return params

    def _get_espell_params(self, **kargs):
        params = self._get_params([], **kargs)
        return params

    def _get_efetch_params(self, **kargs):
        keys = ['WebEnv', 'query_key', 'retmode', 'rettype', 'retstart',
                'retmax', 'strand', 'seq_start', 'seq_stop', 'complexity']
        params = self._get_params(keys, **kargs)
        return params

    def _get_elink_params(self, **kargs):
        # Note that id could be id[] ?
        keys = ['reldate', 'mindate', 'maxdate', 'datetype',
                'term', 'holding', 'linkname', 'WebEnv', 'query_key', 'cmd']
        params = self._get_params(keys, **kargs)
        return params

    def _get_epost_params(self, **kargs):
        params = self._get_params(['WebEnv'], **kargs)
        return params

    def _check_ids(self, sid):
        if sid is None:
            return sid
        elif isinstance(sid, int):
            sid = str(sid)
        elif isinstance(sid, list):
            sid = ",".join([str(x) for x in sid])

        # If there are commas, let us split, strip spaces and join back the ids
        sid = ",".join([x.strip() for x in sid.split(',') if x.strip()!=""])

        if len(sid.split(",")) > 200:
            raise ValueError("Number of comma separated IDs must be less than 200")
        return sid

    def taxonomy_summary(self, id):
        """Alias to EFetch for the taxonomy database

        ::

            >>> s = EUtils()
            >>> ret = s.taxonomy("9606")
            >>> ret['9606']['species']
            'sapiens'
            >>> ret = s.taxonomy("9606,9605,111111111,9604")
            >>> ret['9604']['taxid']
            9604

        """
        sid = self._check_ids(id)
        ret = self.ESummary('taxonomy', sid)
        return ret

    def snp_summary(self, id):
        """Alias to Efetch for the SNP database


        :Return: a json data structure.

        ::

            >>> ret = s.snp("123")


        """
        ret = self.ESummary("snp", id)
        return ret

    def EFetch(self, db, id, retmode="text", **kargs):
        """Access to the EFetch E-Utilities

        :param str db: database from which to retrieve UIDs.
        :param str id: list of identifiers.
        :param retmode: default to text (could be xml but not recommended).
        :param rettype: could be fasta, summary, docsum
        :return: depends on retmode parameter.


        .. note:: addition to NCBI: settings rettype to "dict" returns a dictionary
        ::

            >>> ret = s.EFetch("omim", "269840")  --> ZAP70
            >>> ret = s.EFetch("taxonomy", "9606", retmode="xml")
            >>> [x.text for x in ret.getchildren()[0].getchildren() if x.tag=="ScientificName"]
            ['Homo sapiens']

            >>> s = eutils.EUtils()
            >>> s.EFetch("protein", "34577063", retmode="text", rettype="fasta")
            >gi|34577063|ref|NP_001117.2| adenylosuccinate synthetase isozyme 2 [Homo sapiens]
            MAFAETYPAASSLPNGDCGRPRARPGGNRVTVVLGAQWGDEGKGKVVDLLAQDADIVCRCQGGNNAGHTV
            VVDSVEYDFHLLPSGIINPNVTAFIGNGVVIHLPGLFEEAEKNVQKGKGLEGWEKRLIISDRAHIVFDFH
            QAADGIQEQQRQEQAGKNLGTTKKGIGPVYSSKAARSGLRMCDLVSDFDGFSERFKVLANQYKSIYPTLE
            IDIEGELQKLKGYMEKIKPMVRDGVYFLYEALHGPPKKILVEGANAALLDIDFGTYPFVTSSNCTVGGVC
            TGLGMPPQNVGEVYGVVKAYTTRVGIGAFPTEQDNEIGELLQTRGREFGVTTGRKRRCGWLDLVLLKYAH
            MINGFTALALTKLDILDMFTEIKVGVAYKLDGEIIPHIPANQEVLNKVEVQYKTLPGWNTDISNARAFKE
            LPVNAQNYVRFIEDELQIPVKWIGVGKSRESMIQLF


        Identifiers could be provided as a single string with comma-separated
        values, or a list of strings, a list of integers, or just one
        string or one integer but no mixing of types in the list::

            >>> e.EFetch("protein", "352, 234", retmode="text", rettype="fasta")
            >>> e.EFetch("protein", 352, retmode="text", rettype="fasta")
            >>> e.EFetch("protein", [352], retmode="text", rettype="fasta")
            >>> e.EFetch("protein", [352, 234], retmode="text", rettype="fasta")


        **retmode** should be xml or text depending on the database.
        For instance, xml for pubmed::

            >>> e.EFetch("pubmed", "20210808", retmode="xml")
            >>> e.EFetch('nucleotide', id=15, retmode='xml')
            >>> e.EFetch('nucleotide', id=15, retmode='text', rettype='fasta')
            >>> e.EFetch('nucleotide', 'NT_019265', rettype='gb')

        Other special characters, such as quotation marks (") or the # symbol
        used in referring to a query key on the History server, should be
        represented by their URL encodings (%22 for "; %23 for #).


        A useful command is the following one that allows to get back a GI
        identifier from its accession, which is common to NCBI/EMBL::

            e.EFetch(db="nuccore",id="AP013055", rettype="seqid", retmode="text")

        .. versionchanged:: 1.5.0
            instead of "xml", retmode can now be set to dict, in which case an
            XML is retrieved and converted to a dictionary if possible.

        """
        _retmode = retmode[:]

        if retmode == "dict":
            retmode = "xml"

        self._check_db(db)
        #self._check_retmode(retmode, valids=['text', 'xml'])
        sid = self._check_ids(id)

        params = self._get_efetch_params(**kargs)

        if 'strand' in params.keys() and params['strand'] != None:
            self.devtools.check_param_in_list(params['strand'], [1, 2])
        if 'complexity' in params.keys() and params['complexity'] != None:
            self.devtools.check_param_in_list(params['complexity'],
                    [0, 1, 2, 3, 4])

        query = "efetch.fcgi?db=%s&id=%s&retmode=%s" % (db, sid, retmode)

        ret = self.http_get(query, params=params)
        try: ret = ret.content
        except: pass

        if _retmode == "dict" and isinstance(ret, (bytes,str)):
            ret = self.parse_xml(ret, "dict")

        return ret

    def EInfo(self, db=None, **kargs):
        """Provides information about a database (e.g., number of records)

        :param str db: target database about which to gather statistics.
            Value must be a valid Entrez database name. See :attr:`databases`
            or don't provide any value to obtain the entire list
        :return: a json data structure that depends on the
           value of :attr:`databases` (default to json)

        ::

            >>> all_database_names = s.EInfo()
            >>> # specific info about one database:
            >>> ret = s.EInfo("taxonomy")
            >>> ret[0]['count']
            u'1445358'
            >>> ret = s.EInfo('pubmed')
            >>> ret[0]['fieldlist'][2]['fullname']
            'Filter'

        You can use the *retmode* parameter to 'xml' as well. In that
        case, you will need a XML parser.

        ::

            >>> ret = s.EInfo("taxonomy")

        .. note:: Note that the name in the XML or json outputs
            differ (some have lower cases, some have upper cases). This
            is inherent to the output of EUtils.

        """
        if db is not None:
            self._check_db(db)
        else:
            return self.databases

        kargs['retmode'] = "json"

        # let us create the query now
        query = 'einfo.fcgi'
        if db is not None:
            query += '?db=%s' % db

        # with parameters
        params = self._get_einfo_params(**kargs)

        # the real call using GET method
        ret = self.http_get(query,  frmt="json", params=params)
        try: ret = ret.content
        except: pass

        try:
            return ret['einforesult']['dbinfo']
        except:
            return ret

    def parse_xml(self, ret, method=None):
        if method is None:
            method = self._xmlparser

        if method == 'EUtilsParser':
            ret = self.easyXML(ret)
            return EUtilsParser(ret)
        elif method == 'objectify': # used in docstrings
            from bioservices.xmltools import XMLObjectify
            return XMLObjectify(ret)
        elif method == 'dict':
            import xmltodict
            return xmltodict.parse(ret)

    def ESummary(self, db, id=None,  **kargs):
        """Returns document summaries for a list of input UIDs


        :param db: a valid database
        :param str id: list of identifiers (or string comma separated).
            all of the UIDs must be from the database specified by db. Limited
            to 200 identifiers

        ::

            >>> from bioservices import *
            >>> s = EUtils()
            >>> ret = s.ESummary("snp","7535")
            >>> ret = s.ESummary("snp","7535,7530")
            >>> ret = s.ESummary("taxonomy", "9606,9913")

        ::

            >>> proteins = e.ESearch("protein", "bacteriorhodopsin",
                    retmax=20)
            >>> ret = e.ESummary("protein", 449301857)
            >>> ret['result']['449301857']['extra']
            'gi|449301857|gb|EMC97866.1||gnl|WGS:AEIF|BAUCODRAFT_31870'


        """
        sid = self._check_ids(id)
        self._check_db(db)
        kargs['retmode'] = "json"

        params = self._get_esummary_params(**kargs)
        # the real call using GET method
        query = "esummary.fcgi?db=%s&id=%s" % (db, sid)
        ret = self.http_get(query, frmt="json",  params=params)
        try:
            return ret['result']
        except:
            return ret

    def EGQuery(self, term, **kargs):
        """Provides the number of records retrieved in all Entrez databases by a text query.

        :param str term: Entrez text query.
            Spaces may be replaced by '+' signs. For very long queries
            (more than  several hundred characters long), consider using
            an HTTP POST call. See the
            PubMed or Entrez help for information about search field
            descriptions and tags.
            Search fields and tags are database specific.
        :return: returns a json data structure

        ::

            >>> ret = s.EGQuery("asthma")
            >>> [(x.DbName, x.Count) for x in ret.eGQueryResult.ResultItem if x.Count!='0']

            >>> ret = s.EGQuery("asthma")
            >>> ret.eGQueryResult.ResultItem[0]
            {'Count': '115241',
             'DbName': 'pmc',
             'MenuName': 'PubMed Central',
             'Status': 'Ok'}


        """
        params = self._get_egquery_params(**kargs)

        query = "egquery.fcgi?term=%s" % (term)
        ret = self.http_get(query, frmt="xml",  params=params)
        try:
            ret = self.parse_xml(ret)['Result']
            return ret
        except:
            return ret

    def ESearch(self, db, term, **kargs):
        """Responds to a query in a given database

        The response can be used later in ESummary, EFetch or ELink,
        along with the term translations of the query.

        :param db: a valid database
        :param term: an Entrez text query

        .. note:: see :meth:`_get_esearch_params` for the list of valid parameters.

        ::

            >>> ret = e.ESearch('protein', 'human', RetMax=5)
            >>> ret = e.ESearch('taxonomy', 'Staphylococcus aureus[all names]')
            >>> ret = e.ESearch('pubmed', "cokelaer AND BioServices")

            >>> ret = e.ESearch('protein', '15718680')
            >>> # Let us show the first pubmed identifier in a browser
            >>> identifiers = e.pubmed(ret['idlist'][0])

        More complex requests can be used. We will not cover all the
        possiblities (see the NCBI website). Here is an example to tune
        the search term to look into PubMed for the journal PNAS
        Volume 16, and retrieve.::

            >>> e.ESearch("pubmed", "PNAS[ta] AND 16[vi]")

        You can then look more closely at a specific identifier using EFetch::

            >>> e = EFetch("pubmed")
            >>> e.Efetch(identifiers)

        .. note:: valid parameters can be found by calling
            :meth:`_get_esearch_params`
        """
        self._check_db(db)
        kargs['retmode'] = "json"

        params = self._get_esearch_params(**kargs)

        query = "esearch.fcgi?db=%s&term=%s" % (db, term)
        ret = self.http_get(query, frmt="json",  params=params)
        try:
            return ret['esearchresult']
        except:
            return ret

    def ESpell(self, db, term, **kargs):
        """Retrieve spelling suggestions for a text query in a given database.

        :param str db: database to search. Value must be a valid Entrez
            database name (default = pubmed).
        :param str term: Entrez text query. All special characters must be
            URL encoded.

        ::

            >>> ret = e.ESpell(db="pubmed", term="aasthma+OR+alergy")
            >>> ret = ret['eSpellResult']
            >>> ret['Query']            'asthmaa OR alergies'
            >>> ret['CorrectedQuery']
            'asthma or allergy'
            >>> ret = e.ESpell(db="pubmed", term="biosservices")
            >>> ret = ret['eSpellResult']
            >>> ret['CorrectedQuery']
            bioservices

        """
        self._check_db(db)

        params = self._get_esearch_params(**kargs)

        query = "espell.fcgi?db=%s&term=%s" % (db, term)
        ret = self.http_get(query, frmt="json",  params=params)
        try: 
            ret = ret.content
            ret = self.parse_xml(ret, 'EUtilsParser')
            return ret
        except: 
            return ret


    def ECitMatch(self, bdata, **kargs):
        r"""


        :param bdata: Citation strings. Each input citation must
            be represented by a citation string in the following format::

                journal_title|year|volume|first_page|author_name|your_key|

            Multiple citation strings may be provided by separating the
            strings with a carriage return character (%0D) or simply \\r or \\n.

            The your_key value is an arbitrary label provided by the user
            that may serve as a local identifier for the citation,
            and it will be included in the output.

            all spaces must be replaced by + symbols and that citation
            strings should end with a final vertical bar |.


        Only xml supported at the time of this implementation.

        ::

            from bioservices import EUtils
            s = EUtils()
            print(s.ECitMatch("proc+natl+acad+sci+u+s+a|1991|88|3248|mann+bj|Art1|%0Dscience|1987|235|182|palmenberg+ac|Art2|"))

        """
        # Fixes https://github.com/cokelaer/bioservices/issues/169
        from urllib.parse import unquote
        params = {'bdata': unquote(bdata), "retmode": "xml"}

        # note here, we use .cgi not .fcgi
        query = "ecitmatch.cgi?db=pubmed&retmode=xml"
        ret = self.http_get(query, None,  params=params)
        try: ret = ret.content
        except: pass

        return ret

    def ELink(self, db=None, dbfrom=None, id=None, **kargs):
        """The Entrez links utility

        Responds to a list of UIDs in a given database with either a list of
        related UIDs (and relevancy scores) in the same database or a list
        of linked UIDs in another Entrez database;

        :param str db: valid database from which to retrieve UIDs.
        :param str dbfrom: Database containing the input UIDs. The
            value must be a valid database name (default = pubmed).
            This is the origin database of
            the link operation. If db and dbfrom are set to the same database
            value, then  ELink will return computational neighbors within
            that database. Computational neighbors have linknames that begin
            with dbname_dbname (examples: protein_protein,
            pcassay_pcassay_activityneighbor).
        :param str id: UID list. Either a single UID or a comma-delimited list
            Limited  to 200 Ids
        :param str cmd: ELink command mode. The command mode specified which
            function ELink will perform. Some optional parameters only
            function for certain values of cmd (see
            http://www.ncbi.nlm.nih.gov/books/NBK25499/#chapter4.ELink).
            Examples are neighbor, prlinks.

        ::

            >>> # Example: Find related articles to PMID 20210808
            >>> ret = s.ELink("pubmed", id="20210808", cmd="neighbor_score")

            >>> ret = s.parse_xml(ret, 'EUtilsParser')
            >>> ret.eLinkResult.LinkSet.LinkSetDb[0].Link[1]
            {'Id': '16539535'}


            >>> s.ELink(dbfrom="nucleotide", db="protein",
                              id="48819,7140345")
            >>> s.ELink(dbfrom="nucleotide", db="protein",
                              id="48819,7140345")
            >>> s.ELink(dbfrom='nuccore', id='21614549,219152114',
                    cmd='ncheck')

        Convert GI number to Taxon identifiers::

            >>> s.ELink(dbfrom='nuccore', db="taxonomy", id='21614549,219152114')



        """
        # unlike other EUtils, db and dbfrom are here optional
        sid = self._check_ids(id)
        if db is not None:
            self._check_db(db)
        if dbfrom is not None:
            self._check_db(dbfrom)
        if db is None and dbfrom is None:
            raise ValueError("One of db or dbfrom parameter must be provided")

        if 'cmd' in kargs.keys():
            assert kargs['cmd'] in ["neighbor", "neighbor_score",
                    "neighbor_history", "acheck", "llinks", "lcheck",
                    "ncheck", "llinkslib", "prlinks"]
            cmd = kargs['cmd']
        else:
            cmd = None

        if db is not None and dbfrom is not None:
            query = "elink.fcgi?db=%s&dbfrom=%s" % (db, dbfrom)
        elif dbfrom is not None:
            query = "elink.fcgi?dbfrom=%s" % dbfrom
        elif db is not None:
            query = "elink.fcgi?db=%s" % db

        if sid is not None:
            query += "&id=%s" % sid
        if cmd is not None:
            query += "&cmd=%s" % cmd

        params = self._get_elink_params(**kargs)

        ret = self.http_get(query, frmt="txt",  params=params)
        #try: ret = ret.content
        #except: pass

        return ret

    def EPost(self, db, id, **kargs):
        """Accepts a list of UIDs from a given database,

        stores the set on the History Server, and responds with a query
        key and web environment for the uploaded dataset.

        :param str db: a valid database
        :param id: list of strings of strings

        :return: a dictionary with a Web Environment string
            and a QueryKey to be re-used in another EUtils.
        """
        self._check_db(db)
        sid = self._check_ids(id)

        params = self._get_epost_params(**kargs)

        query = "epost.fcgi/?db=%s&id=%s" % (db, sid)

        ret = self.http_get(query, "xml",  params=params)
        try: ret = ret.content
        except: pass
        ret = self.easyXML(ret)
        for item in ret.getchildren():
            if item.tag == 'QueryKey':
                query_key = item.text
            elif item.tag == 'WebEnv':
                webenv = item.text
        return {'WebEnv':webenv, 'QueryKey':query_key}



class AttrDict(dict):
    def __init__(self, *args, **kwargs):
        super(AttrDict, self).__init__(*args, **kwargs)
        self.__dict__ = self


class EUtilsParser(AttrDict):
    """Convert xml returned by EUtils into a structure easier to manipulate

    Used by :meth:`EUtils.EGQuery`, :meth:`EUtils.ELink`.
    """
    def __init__(self, xml):
        super(EUtilsParser, self).__init__()

        try:
            name = xml.root.tag
            self[name] = EUtilsParser(xml.root)
            children = []
            #children = xml.root.getchildren()[0].getchildren()
            #self.__name = xml.root.getchildren()[0].tag
        except:
            children = [x for x in xml]
            if len(children) == 0:
                self[xml.tag] = xml.text

        for i, child in enumerate(children):
            if len([x for x in child]) == 0:
                if child.tag in self.keys():
                    try:
                        self[child.tag].append(child.text)
                    except:
                        self[child.tag] = [self[child.tag]]
                        self[child.tag].append(child.text)
                else:
                    self[child.tag] = child.text
            else:
                # This is probably a list then
                e = EUtilsParser(child)
                if child.tag not in self.keys():
                    self[child.tag] = e
                else:
                    try:
                        self[child.tag].append(e)
                    except:
                        self[child.tag] = [self[child.tag]]
                        self[child.tag].append(e)


    def __str__(self):
        name = self._EUtilsParser__name
        if name == "DbInfo":
            txt = ""
            for this in self.FieldList:
                txt += "{0:10}:{1}\n".format(this.Name, this.Description)
            return txt
        else:
            print("Not implemented for {0}".format(name))




class XMLEUtils(object):
    def __init__(self, xml):
        self.xml = xml





#
#  This file is part of bioservices software
#
#  Copyright (c) 2013-2014 - EBI-EMBL
#
#  File author(s):
#      Thomas Cokelaer <cokelaer@ebi.ac.uk>
#
#
#  Distributed under the GPLv3 License.
#  See accompanying file LICENSE.txt or copy at
#      http://www.gnu.org/licenses/gpl-3.0.html
#
#  website: https://github.com/cokelaer/bioservices
#  documentation: http://packages.python.org/bioservices
#
##############################################################################
"""Interface to some part of the UniProt web service

.. topic:: What is UniProt ?

    :URL: http://www.uniprot.org
    :Citation:

    .. highlights::

        "The Universal Protein Resource (UniProt) is a comprehensive resource for protein
        sequence and annotation data. The UniProt databases are the UniProt
        Knowledgebase (UniProtKB), the UniProt Reference Clusters (UniRef), and the
        UniProt Archive (UniParc). The UniProt Metagenomic and Environmental Sequences
        (UniMES) database is a repository specifically developed for metagenomic and
        environmental data."

        -- From Uniprot web site (help/about) , Dec 2012


.. mapping between uniprot and bunch of other DBs.
.. ftp://ftp.uniprot.org/pub/databases/uniprot/current_release/knowledgebase/idmapping/
.. http://www.uniprot.org/docs/speclist
.. http://www.uniprot.org/docs/pkinfam

"""
import types
import io
import sys

from bioservices.services import REST
from bioservices import logger
logger.name = __name__

try:
    import pandas as pd
except:
    pass
__all__ = ["UniProt"]

# TODO:: falt files to get list of identifiers
# http://www.ebi.ac.uk/uniprot/database/download.html
# grep sp uniprot_sprot.fasta  | grep HUMAN | awk '{print substr($1, 12, length($1))}'

mapping = {"UniProtKB AC/ID": "ACC+ID",
           "UniProtKB": "ACC",
           "UniProtKB": "ID",
           "UniParc": "UPARC",
           "UniRef50": "NF50",
           "UniRef90": "NF90",
           "UniRef100": "NF100",
           "EMBL/GenBank/DDBJ": "EMBL_ID",
           "EMBL/GenBank/DDBJ CDS": "EMBL",
           "PIR": "PIR",
           "UniGene": "UNIGENE_ID",
           "Entrez Gene (GeneID)": "P_ENTREZGENEID",
           "GI number*": "P_GI",
           "IPI": "P_IPI",
           "RefSeq Protein": "P_REFSEQ_AC",
           "RefSeq Nucleotide": "REFSEQ_NT_ID",
           "PDB": "PDB_ID",
           "DisProt": "DISPROT_ID",
           "HSSP": "HSSP_ID",
           "DIP": "DIP_ID",
           "MINT": "MINT_ID",
           "Allergome": "ALLERGOME_ID",
           "MEROPS": "MEROPS_ID",
           "mycoCLAP": "MYCOCLAP_ID",
           "PeroxiBase": "PEROXIBASE_ID",
           "PptaseDB": "PPTASEDB_ID",
           "REBASE": "REBASE_ID",
           "TCDB": "TCDB_ID",
           "PhosSite": "PHOSSITE_ID",
           "DMDM": "DMDM_ID",
           "Aarhus/Ghent-2DPAGE": "AARHUS_GHENT_2DPAGE_ID",
           "World-2DPAGE": "WORLD_2DPAGE_ID",
           "DNASU": "DNASU_ID",
           "Ensembl": "ENSEMBL_ID",
           "Ensembl Protein": "ENSEMBL_PRO_ID",
           "Ensembl Transcript": "ENSEMBL_TRS_ID",
           "Ensembl Genomes": "ENSEMBLGENOME_ID",
           "Ensembl Genomes Protein": "ENSEMBLGENOME_PRO_ID",
           "Ensembl Genomes Transcript": "ENSEMBLGENOME_TRS_ID",
           "GeneID": "P_ENTREZGENEID",
           "GenomeReviews": "GENOMEREVIEWS_ID",
           "KEGG": "KEGG_ID",
           "PATRIC": "PATRIC_ID",
           "UCSC": "UCSC_ID",
           "VectorBase": "VECTORBASE_ID",
           "AGD": "AGD_ID",
           "ArachnoServer": "ARACHNOSERVER_ID",
           "CGD": "CGD",
           "ConoServer": "CONOSERVER_ID",
           "CYGD": "CYGD_ID",
           "dictyBase": "DICTYBASE_ID",
           "EchoBASE": "ECHOBASE_ID",
           "EcoGene": "ECOGENE_ID",
           "euHCVdb": "EUHCVDB_ID",
           "EuPathDB": "EUPATHDB_ID",
           "FlyBase": "FLYBASE_ID",
           "GeneCards": "GENECARDS_ID",
           "GeneFarm": "GENEFARM_ID",
           "GenoList": "GENOLIST_ID",
           "H-InvDB": "H_INVDB_ID",
           "HGNC": "HGNC_ID",
           "HPA": "HPA_ID",
           "LegioList": "LEGIOLIST_ID",
           "Leproma": "LEPROMA_ID",
           "MaizeGDB": "MAIZEGDB_ID",
           "MIM": "MIM_ID",
           "MGI": "MGI_ID",
           "neXtProt": "NEXTPROT_ID",
           "Orphanet": "ORPHANET_ID",
           "PharmGKB": "PHARMGKB_ID",
           "PomBase": "POMBASE_ID",
           "PseudoCAP": "PSEUDOCAP_ID",
           "RGD": "RGD_ID",
           "SGD": "SGD_ID",
           "TAIR": "TAIR_ID",
           "TubercuList": "TUBERCULIST_ID",
           "WormBase": "WORMBASE_ID",
           "WormBase Transcript": "WORMBASE_TRS_ID",
           "WormBase Protein": "WORMBASE_PRO_ID",
           "Xenbase": "XENBASE_ID",
           "ZFIN": "ZFIN_ID",
           "eggNOG": "EGGNOG_ID",
           "GeneTree": "GENETREE_ID",
           "HOGENOM": "HOGENOM_ID",
           "HOVERGEN": "HOVERGEN_ID",
           "KO": "KO_ID",
           "OMA": "OMA_ID",
           "OrthoDB": "ORTHODB_ID",
           "ProtClustDB": "PROTCLUSTDB_ID",
           "BioCyc": "BIOCYC_ID",
           "Reactome": "REACTOME_ID",
           "UniPathWay": "UNIPATHWAY_ID",
           "CleanEx": "CLEANEX_ID",
           "GermOnline": "GERMONLINE_ID",
           "ChEMBL": "CHEMBL_ID",
           "ChiTaRS": "CHITARS_ID",
           "DrugBank": "DRUGBANK_ID",
           "GenomeRNAi": "GENOMERNAI_ID",
           "NextBio": "NEXTBIO_ID"}


class UniProt(REST):
    """Interface to the `UniProt <http://www.uniprot.org>`_ service

    .. rubric:: Identifiers mapping between databases:

    ::

        >>> from bioservices import UniProt
        >>> u = UniProt(verbose=False)
        >>> u.mapping("ACC", "KEGG_ID", query='P43403')
        defaultdict(<type 'list'>, {'P43403': ['hsa:7535']})
        >>> res = u.search("P43403")

        # Returns sequence on the ZAP70_HUMAN accession Id
        >>> sequence = u.search("ZAP70_HUMAN", columns="sequence")

    """
    _mapping = mapping.copy()
    _url = "https://www.uniprot.org"
    # _valid_columns = ['citation', 'clusters', 'comments', 'database',
    #                   'domains', 'domain', 'ec', 'id', 'entry name', 'existence',
    #                   'families', 'feature', 'features', 'genes', 'go', 'go-id', 'interpro',
    #                   'interactor', 'keywords', 'keyword-id', 'last-modified',
    #                   'length', 'organism', 'organism-id', 'pathway', 'protein names',
    #                   'reviewed', 'score', 'sequence', '3d', 'subcellular locations',
    #                   'taxonomy', 'tools', 'version', 'virus hosts', 'lineage-id',
    #                   'sequence-modified', 'proteome']
    _valid_columns = [
        # Names & Taxonomy
        'id', 'entry name', 'genes', 'genes(PREFERRED)', 'genes(ALTERNATIVE)',
        'genes(OLN)', 'genes(ORF)', 'organism', 'organism-id', 'protein names',
        'proteome', 'lineage(ALL)', 'lineage-id', 'virus hosts',
        # Sequences
        'fragment', 'sequence', 'length', 'mass', 'encodedon',
        'comment(ALTERNATIVE PRODUCTS)', 'comment(ERRONEOUS GENE MODEL PREDICTION)',
        'comment(ERRONEOUS INITIATION)', 'comment(ERRONEOUS TERMINATION)',
        'comment(ERRONEOUS TRANSLATION)', 'comment(FRAMESHIFT)',
        'comment(MASS SPECTROMETRY)', 'comment(POLYMORPHISM)',
        'comment(RNA EDITING)', 'comment(SEQUENCE CAUTION)',
        'feature(ALTERNATIVE SEQUENCE)', 'feature(NATURAL VARIANT)',
        'feature(NON ADJACENT RESIDUES)',
        'feature(NON STANDARD RESIDUE)', 'feature(NON TERMINAL RESIDUE)',
        'feature(SEQUENCE CONFLICT)', 'feature(SEQUENCE UNCERTAINTY)',
        'version(sequence)',
        # Family and Domains
        'domains', 'domain', 'comment(DOMAIN)', 'comment(SIMILARITY)',
        'feature(COILED COIL)', 'feature(COMPOSITIONAL BIAS)',
        'feature(DOMAIN EXTENT)', 'feature(MOTIF)', 'feature(REGION)',
        'feature(REPEAT)', 'feature(ZINC FINGER)',
        # Function
        'ec', 'comment(ABSORPTION)', 'comment(CATALYTIC ACTIVITY)',
        'comment(COFACTOR)', 'comment(ENZYME REGULATION)', 'comment(FUNCTION)',
        'comment(KINETICS)', 'comment(PATHWAY)', 'comment(REDOX POTENTIAL)',
        'comment(TEMPERATURE DEPENDENCE)', 'comment(PH DEPENDENCE)',
        'feature(ACTIVE SITE)', 'feature(BINDING SITE)', 'feature(DNA BINDING)',
        'feature(METAL BINDING)', 'feature(NP BIND)', 'feature(SITE)',
        # Gene Ontologys
        'go', 'go(biological process)', 'go(molecular function)',
        'go(cellular component)', 'go-id',
        # InterPro
        'interpro',
        # Interaction
        'interactor', 'comment(SUBUNIT)',
        # Publications
        'citation', 'citationmapping',
        # Date of
        'created', 'last-modified', 'sequence-modified', 'version(entry)',
        # Structure
        '3d', 'feature(BETA STRAND)', 'feature(HELIX)', 'feature(TURN)',
        # Subcellular location
        'comment(SUBCELLULAR LOCATION)', 
        'feature(INTRAMEMBRANE)',
        'feature(TOPOLOGICAL DOMAIN)',
        'feature(TRANSMEMBRANE)',
        # Miscellaneous
        'annotation score', 'score', 'features', 'comment(CAUTION)',
        'comment(TISSUE SPECIFICITY)',
        'comment(GENERAL)', 'keywords', 'context', 'existence', 'tools',
        'reviewed', 'feature', 'families', 'subcellular locations', 'taxonomy',
        'version', 'clusters', 'comments', 'database', 'keyword-id', 'pathway',
        'score',
        # Pathology & Biotech
        'comment(ALLERGEN)', 'comment(BIOTECHNOLOGY)', 'comment(DISRUPTION PHENOTYPE)',
        'comment(DISEASE)', 'comment(PHARMACEUTICAL)', 'comment(TOXIC DOSE)',
        # PTM / Processsing
        'comment(PTM)', 'feature(CHAIN)', 'feature(CROSS LINK)', 'feature(DISULFIDE BOND)',
        'feature(GLYCOSYLATION)', 'feature(INITIATOR METHIONINE)', 'feature(LIPIDATION)',
        'feature(MODIFIED RESIDUE)', 'feature(PEPTIDE)', 'feature(PROPEPTIDE)',
        'feature(SIGNAL)', 'feature(TRANSIT)',
        # Taxonomic lineage
        'lineage(all)', 'lineage(SUPERKINGDOM)', 'lineage(KINGDOM)', 'lineage(SUBKINGDOM)',
        'lineage(SUPERPHYLUM)', 'lineage(PHYLUM)', 'lineage(SUBPHYLUM)', 'lineage(SUPERCLASS)',
        'lineage(CLASS)', 'lineage(SUBCLASS)', 'lineage(INFRACLASS)', 'lineage(SUPERORDER)',
        'lineage(ORDER)', 'lineage(SUBORDER)', 'lineage(INFRAORDER)', 'lineage(PARVORDER)',
        'lineage(SUPERFAMILY)', 'lineage(FAMILY)', 'lineage(SUBFAMILY)', 'lineage(TRIBE)',
        'lineage(SUBTRIBE)', 'lineage(GENUS)', 'lineage(SUBGENUS)', 'lineage(SPECIES GROUP)',
        'lineage(SPECIES SUBGROUP)', 'lineage(SPECIES)', 'lineage(SUBSPECIES)', 'lineage(VARIETAS)',
        'lineage(FORMA)',
        # Taxonomic identifier
        'lineage-id(all)', 'lineage-id(SUPERKINGDOM)', 'lineage-id(KINGDOM)', 'lineage-id(SUBKINGDOM)',
        'lineage-id(SUPERPHYLUM)', 'lineage-id(PHYLUM)', 'lineage-id(SUBPHYLUM)', 'lineage-id(SUPERCLASS)',
        'lineage-id(CLASS)', 'lineage-id(SUBCLASS)', 'lineage-id(INFRACLASS)', 'lineage-id(SUPERORDER)',
        'lineage-id(ORDER)', 'lineage-id(SUBORDER)', 'lineage-id(INFRAORDER)', 'lineage-id(PARVORDER)',
        'lineage-id(SUPERFAMILY)', 'lineage-id(FAMILY)', 'lineage-id(SUBFAMILY)', 'lineage-id(TRIBE)',
        'lineage-id(SUBTRIBE)', 'lineage-id(GENUS)', 'lineage-id(SUBGENUS)', 'lineage-id(SPECIES GROUP)',
        'lineage-id(SPECIES SUBGROUP)', 'lineage-id(SPECIES)', 'lineage-id(SUBSPECIES)', 'lineage-id(VARIETAS)',
        'lineage-id(FORMA)',
        # Cross-references
        'database(db_abbrev)', 'database(EMBL)']

    def __init__(self, verbose=False, cache=False):
        """**Constructor**

        :param verbose: set to False to prevent informative messages
        """
        super(UniProt, self).__init__(name="UniProt", url=UniProt._url,
                                      verbose=verbose, cache=cache)
        self.TIMEOUT = 100

        self._database = "uniprot"

    def _download_flat_files(self):
        """could be used to get all data in flat files (about compressed 500Mb )"""
        url = "ftp://ftp.ebi.ac.uk/pub/databases/uniprot/knowledgebase/uniprot_sprot.dat.gz"
        self.logging.info('Downloading uniprot file from the web. May take some time.:')
        import urllib
        urllib.urlretrieve(url, 'uniprot_sprot.dat.gz')

    def mapping(self, fr="ID", to="KEGG_ID", query="P13368"):
        """This is an interface to the UniProt mapping service

        :param fr: the source database identifier. See :attr:`_mapping`.
        :param to: the targetted database identifier. See :attr:`_mapping`.
        :param query: a string containing one or more IDs separated by a space
            It can also be a list of strings.
        :param format: The output being a dictionary, this parameter is
            deprecated and not used anymore
        :return: a list. The first element is the source database Id. The second
            is the targetted source identifier. Following elements are alternate
            of one the entry and its mapped Id. If a query has several mapped
            Ids, the query is repeated (see example with PDB mapping here below)
            e.g., ["From:ID", "to:PDB_ID", "P43403"]

        ::

            >>> u.mapping("ACC", "KEGG_ID", 'P43403')
            defaultdict(<type 'list'>, {'P43403': ['hsa:7535']})
            >>> u.mapping("ACC", "KEGG_ID", 'P43403 P00958')
            defaultdict(<type 'list'>, {'P00958': ['sce:YGR264C'], 'P43403': ['hsa:7535']})
            >>> u.mapping("ID", "PDB_ID", "P43403")
            defaultdict(<type 'list'>, {'P43403': ['1FBV', '1M61', '1U59',
            '2CBL', '2OQ1', '2OZO', '2Y1N', '3ZNI', '4A4B', '4A4C', '4K2R']})

        There is a web page that gives the list of correct `database identifiers
        <http://www.uniprot.org/faq/28>`_. You can also look at the
        :attr:`_mapping` attribute.

        :URL: http://www.uniprot.org/mapping/

        .. versionchanged:: 1.1.1 to return a dictionary instaed of a list
        .. versionchanged:: 1.1.2 the values for each key is now made of a list
            instead of strings so as to store more than one values.
        .. versionchanged:: 1.2.0 input query can also be a list of strings
            instead of just a string
        .. versionchanged:: 1.3.1:: use http_post instead of http_get. This is 3 times
            faster and allows queries with more than 600 entries in one go.
        """
        url = 'mapping/'  # the slash matters

        query = self.devtools.list2string(query, sep=" ", space=False)
        # if isinstance(query, list):
        #    query = " ".join(query)
        params = {'from': fr, 'to': to, 'format': "tab", 'query': query}
        result = self.http_post(url, frmt="txt", data=params)

        # changes in version 1.1.1 returns a dictionary instead of list
        try:
            result = result.split()
            del result[0]
            del result[0]
        except:
            self.logging.warning("Results seems empty...returning empty dictionary.")
            return {}

        if len(result) == 0:
            return {}
        else:
            # bug fix based on ticket #19 version 1.1.2
            # the default dict set empty list for all keys by default
            from collections import defaultdict
            result_dict = defaultdict(list)

            keys = result[0::2]
            values = result[1::2]
            for i, key in enumerate(keys):
                result_dict[key].append(values[i])
        return result_dict

    def searchUniProtId(self, uniprot_id, frmt="xml"):
        self.logging.warning("DEPRECATED SINCE VERSION 1.3.1. use retrieve instead")

    def retrieve(self, uniprot_id, frmt="xml", database="uniprot"):
        """Search for a uniprot ID in UniProtKB database

        :param str uniprot: a valid UniProtKB ID or a list of identifiers.
        :param str frmt: expected output format amongst xml, txt, fasta, gff, rdf
        :return: is a list of identifiers is provided, the output is also a list
            otherwise, a string. The content of the string of items in the list
            depends on the value of **frmt**.

        ::

            >>> u = UniProt()
            >>> res = u.retrieve("P09958", frmt="xml")
            >>> fasta = u.retrieve([u'P29317', u'Q5BKX8', u'Q8TCD6'], frmt='fasta')
            >>> print(fasta[0])

        """
        _valid_formats = ['txt', 'xml', 'rdf', 'gff', 'fasta']
        self.devtools.check_param_in_list(frmt, _valid_formats)

        queries = self.devtools.to_list(uniprot_id)

        url = [database + "/" + query + '.' + frmt for query in queries]
        res = self.http_get(url, frmt="txt")
        if frmt == "xml":
            res = [self.easyXML(x) for x in res]
        if isinstance(res, list) and len(res) == 1:
            res = res[0]
        return res

    def get_fasta(self, id_):
        """Returns FASTA string given a valid identifier


        .. seealso:: :mod:`bioservices.apps.fasta` for dedicated tools to
            manipulate FASTA
        """
        from bioservices.apps.fasta import FASTA
        f = FASTA()
        f.load_fasta(id_)
        return f.fasta

    def get_fasta_sequence(self, id_):
        """Returns FASTA sequence (Not FASTA)

        :param str id_: Should be the entry name
        :return: returns fasta sequence (string)

        .. warning:: this is the sequence found in a fasta file, not the fasta
            content itself. The difference is that the header is removed and the
            formatting of end of lines every 60 characters is removed.

        """
        from bioservices.apps.fasta import FASTA
        f = FASTA()
        f.load_fasta(id_)
        return f.sequence

    def search(self, query, frmt="tab", columns=None,
               include=False, sort="score", compress=False, limit=None,
               offset=None, maxTrials=10, database="uniprot"):
        """Provide some interface to the uniprot search interface.

        :param str query: query must be a valid uniprot query.
            See http://www.uniprot.org/help/text-search, http://www.uniprot.org/help/query-fields
            See also example below
        :param str frmt: a valid format amongst html, tab, xls, asta, gff,
            txt, xml, rdf, list, rss. If tab or xls, you can also provide the
            columns argument.  (default is tab)
        :param str columns: comma-separated list of values. Works only if fomat
            is tab or xls. For UnitProtKB, some possible columns are:
            id, entry name, length, organism. Some column name must be followed by
            database name (e.g., "database(PDB)"). Again, see uniprot website
            for more details. See also :attr:`~bioservices.uniprot.UniProt._valid_columns`
            for the full list of column keyword.
        :param bool include: include isoform sequences when the frmt
            parameter is fasta. Include description when frmt is rdf.
        :param str sort: by score by default. Set to None to bypass this behaviour
        :param bool compress: gzip the results
        :param int limit: Maximum number of results to retrieve.
        :param int offset:  Offset of the first result, typically used together
            with the limit parameter.
        :param int maxTrials: this request is unstable, so we may want to try
            several time.

        To obtain the list of uniprot ID returned by the search of zap70 can be
        retrieved as follows::

            >>> u.search('zap70+AND+organism:9606', frmt='list')
            >>> u.search("zap70+and+taxonomy:9606", frmt="tab", limit=3,
            ...    columns="entry name,length,id, genes")
            Entry name  Length  Entry   Gene names
            CBLB_HUMAN  982 Q13191  CBLB RNF56 Nbla00127
            CBL_HUMAN   906 P22681  CBL CBL2 RNF55
            CD3Z_HUMAN  164 P20963  CD247 CD3Z T3Z TCRZ

        other examples::

            >>> u.search("ZAP70+AND+organism:9606", limit=3, columns="id,database(PDB)")

        You can also do a search on several keywords. This is especially useful
        if you have a list of known entry names.::

            >>> u.search("ZAP70_HUMAN+or+CBL_HUMAN", frmt="tab", limit=3,
            ...    columns="entry name,length,id, genes")
            Entry name  Length  Entry   Gene names


        Finally, note that when search for a query, you may have several hits::

            >>> u.search("P12345)

        including the ID P12345 but also related entries. If you need only the
        entry that perfectly match the query, use::

            >>> u.search("id:P12345")

        This was provided from a user issue that was solved here:
        https://github.com/cokelaer/bioservices/issues/122


        .. warning:: some columns although valid may not return anything, not even in
            the header: 'score', 'taxonomy', 'tools'. this is a uniprot feature,
            not bioservices.
        """
        params = {}

        if frmt is not None:
            _valid_formats = ['tab', 'xls', 'fasta', 'gff', 'txt', 'xml', 'rss', 'list', 'rss', 'html']
            self.devtools.check_param_in_list(frmt, _valid_formats)
            params['format'] = frmt

        if columns is not None:
            self.devtools.check_param_in_list(frmt, ["tab", "xls"])

            # remove unneeded spaces before/after commas if any
            if "," in columns:
                columns = [x.strip() for x in columns.split(",")]
            else:
                columns = [columns]

            for col in columns:
                if col.startswith("database(") is True:
                    pass
                else:
                    self.devtools.check_param_in_list(col, self._valid_columns)

            # convert back to a string as expected by uniprot
            params['columns'] = ",".join([x.strip() for x in columns])

        if include is True and frmt in ["fasta", "rdf"]:
            params['include'] = 'yes'

        if compress is True:
            params['compress'] = 'yes'

        if sort:
            self.devtools.check_param_in_list(sort, ["score"])
            params['sort'] = sort

        if offset is not None:
            if isinstance(offset, int):
                params['offset'] = offset

        if limit is not None:
            if isinstance(limit, int):
                params['limit'] = limit

        # + are interpreted and have a meaning.
        params['query'] = query.replace("+", " ")
        # res = s.request("/uniprot/?query=zap70+AND+organism:9606&format=xml", params)
        res = self.http_get(database + "/", frmt="txt", params=params)
        return res

    def quick_search(self, query, include=False, sort="score", limit=None):
        """a specialised version of :meth:`search`

        This is equivalent to::

            u = uniprot.UniProt()
            u.search(query, frmt="tab", include=False, sor="score", limit=None)

        :returns: a dictionary.

        """
        res = self.search(query, "tab", include=include, sort=sort, limit=limit)

        # if empty result, nothing to do
        if res and len(res) == 0:
            return res

        # else populate a dictionary
        newres = {}
        for line in res.split("\n")[1:-1]:
            # print line
            Entry, a, b, c, d, e, f = line.split("\t")
            # print Entry, a, b, c, d, e, f
            newres[Entry] = {'Entry name': a,
                             'Status': b,
                             'Protein names': c,
                             'Gene names': d,
                             'Organism': e,
                             'Length': f}
        return newres

    def uniref(self, query):
        """Calls UniRef service

        :return: if you have Pandas installed, returns a dataframe (see example)

        ::

            >>> u = UniProt()
            >>> df = u.uniref("member:Q03063")  # of just A03063
            >>> df.Size

        Another example from https://github.com/cokelaer/bioservices/issues/121
        is the combination of uniprot and uniref filters::

            u.uniref("uniprot:(ec:1.1.1.282 taxonomy:bacteria reviewed:yes) AND identity:0.5")

        """
        try:
            import pandas as pd
        except:
            print("uniref method requires Pandas", file=sys.stderr)
            return
        res = self.http_get("uniref/", params={"query": query, 'format': 'tab'}, frmt="txt")
        try:
            # python 2.X
            res = pd.read_csv(io.StringIO(unicode(res)), sep="\t")
        except:
            res = pd.read_csv(io.StringIO(str(res.strip())), sep="\t")
        return res

    def get_df(self, entries, nChunk=100, organism=None, limit=10):
        """Given a list of uniprot entries, this method returns a dataframe with all possible columns


        :param entries: list of valid entry name. if list is too large (about
            >200), you need to split the list
        :param chunk:
        :param limit: limit number of entries per identifier to 10. You can 
            set it to None to keep all entries but this will be very slow
        :return: dataframe with indices being the uniprot id (e.g. DIG1_YEAST)

        .. todo:: cleanup the content of the data frame to replace strings
            separated by ; into a list of strings. e.g. the Gene Ontology IDs

        .. warning:: requires pandas library
        """
        if isinstance(entries, str):
            entries = [entries]
        else:
            entries = list(set(entries))
        output = pd.DataFrame()

        self.logging.info("fetching information from uniprot for {} entries".format(len(entries)))

        nChunk = min(nChunk, len(entries))
        N, rest = divmod(len(entries), nChunk)
        for i in range(0, N + 1):
            this_entries = entries[i * nChunk:(i + 1) * nChunk]
            if len(this_entries):
                self.logging.info("uniprot.get_df {}/{}".format(i + 1, N))
                query = "+or+".join(this_entries)
                if organism:
                    query += "+and+" + organism
                res = self.search(query, frmt="tab",
                                  columns=",".join(self._valid_columns),
                                limit=limit)
            else:
                break
            if len(res) == 0:
                self.logging.warning("some entries %s not found" % entries)
            else:
                try:
                    # python 2.X
                    df = pd.read_csv(io.StringIO(unicode(res)), sep="\t")
                except:
                    df = pd.read_csv(io.StringIO(str(res)), sep="\t")
                if isinstance(output, type(None)):
                    output = df.copy()
                else:
                    output = output.append(df, ignore_index=True)

        # you may end up with duplicated...
        output.drop_duplicates(inplace=True)
        # you may have new entries...
        # output = output[output.Entry.apply(lambda x: x in entries)]
        # to transform into list:
        columns = ['PubMed ID', 'Comments', u'Domains', 'Protein families',
                   'Gene names', 'Gene ontology (GO)', 'Gene ontology IDs',
                   'InterPro', 'Interacts with', 'Keywords']
                   #'Subcellular location']
        for col in columns:
            try:
                res = output[col].apply(lambda x: [this.strip() for this in str(x).split(";") if this != "nan"])
                output[col] = res
            except:
                self.logging.warning("column could not be parsed. %s" % col)
        # Sequences are splitted into chunks of 10 characters. let us rmeove
        # the spaces:
        output['Sequence'].fillna("", inplace=True)
        output.Sequence = output['Sequence'].apply(lambda x: x.replace(" ", ""))

        return output



from bioservices.apps import download_fasta as df
import pytest



@pytest.mark.flaky
def test_ena():

    from easydev import TempFile
    with TempFile() as f1:
        df.download_fasta("FN433596.1", output_filename=f1.name, method="ENA")

    with TempFile() as f2:
        df.download_fasta("FN433596.1", output_filename=f2.name, method="EUtils")

    with TempFile() as f3:
        from bioservices import ENA
        ena = ENA()
        df.download_fasta("FN433596.1", output_filename=f2.name, method="EUtils",
            service=ena)

    try:
        df.download_fasta("FN433596.1", method="dummy")
        assert False
    except:
        assert True



# -*- python -*-
#
#  This file is part of bioservices software
#
#  Copyright (c) 2013-2014 - EBI-EMBL
#
#  File author(s):
#      Thomas Cokelaer <cokelaer@ebi.ac.uk>
#
#
#  Distributed under the GPLv3 License.
#  See accompanying file LICENSE.txt or copy at
#      http://www.gnu.org/licenses/gpl-3.0.html
#
#  website: https://github.com/cokelaer/bioservices
#  documentation: http://packages.python.org/bioservices
#
##############################################################################
"""Interface to some part of the UniProt web service

.. topic:: What is UniProt ?

    :URL: http://www.uniprot.org
    :Citation:

    .. highlights::

        "The Universal Protein Resource (UniProt) is a comprehensive resource for protein
        sequence and annotation data. The UniProt databases are the UniProt
        Knowledgebase (UniProtKB), the UniProt Reference Clusters (UniRef), and the
        UniProt Archive (UniParc). The UniProt Metagenomic and Environmental Sequences
        (UniMES) database is a repository specifically developed for metagenomic and
        environmental data."

        -- From Uniprot web site (help/about) , Dec 2012


.. mapping between uniprot and bunch of other DBs.
.. ftp://ftp.uniprot.org/pub/databases/uniprot/current_release/knowledgebase/idmapping/
.. http://www.uniprot.org/docs/speclist
.. http://www.uniprot.org/docs/pkinfam

"""
import re
try:
    from urllib2 import urlopen
except:
    from urllib.request import urlopen

from bs4 import BeautifulSoup
from bioservices.services import REST
from bioservices.xmltools import readXML


__all__ = ["BioCarta"]


# method for unicode transformation
try:
    text = unicode
except NameError:
    text = str


class BioCarta():
    """Interface to `BioCarta <http://www.biocarta.com>`_ pages

    This is not a REST interface actually but rather a parser to some of the
    HTML pages related to pathways.

    One can retrieve the pathways names and their list of proteins.

        >>> from bioservics import *
        >>> b = BioCarta()
        >>> pathways = b.get_pathway_names()
        >>> proteins = b.get_pathway_protein_names(pathways[0])


    .. warning:: biocarta pathways layout can be accesses from PID

    """
    _url = "http://cgap.nci.nih.gov/Pathways/BioCarta_Pathways"

    _organism_prefixes = {'Homo sapiens': 'h', 'Mus musculus': 'm'}
    organisms = set(_organism_prefixes.keys())

    _all_pathways = None
    _pathway_categories = None
    _all_pathways_url =  "http://cgap.nci.nih.gov/Pathways/BioCarta_Pathways"

    def __init__(self, verbose=True, cache=False):
        """**Constructor**

        :param verbose: set to False to prevent informative messages
        """
        self.services = REST(name="BioCarta", url=BioCarta._url, cache=cache,
            verbose=verbose)
        self.fname  = "biocarta_pathways.txt"

        self._organism = None
        self._organism_prefix = None
        self._pathways = None

    # set the default organism used by pathways retrieval
    def _get_organism(self):
        return self._organism

    def _set_organism(self, organism):
        organism = organism[:1].upper() + organism[1:].lower()
        if organism == self._organism: 
            return
        if organism not in BioCarta.organisms:
            raise ValueError("Invalid organism. Check the list in :attr:`organisms` attribute")

        self._organism = organism
        self._organism_prefix = BioCarta._organism_prefixes[organism]
        self._pathways = None

    organism = property(_get_organism, _set_organism, doc="returns the current default organism")

    def _get_pathway_categories(self):
        if self._pathway_categories is None:
            self._pathway_categories = self.services.http_get
        return self._pathway_categories
    pathway_categories = property(_get_pathway_categories)

    def _get_all_pathways(self):
        """returns pathways from biocarta

        human and mouse organisms are available but only those corresponding
        to the organism defined in :attr:`organism` are returned.
        """
        if self.organism is None:
           raise ValueError(
                "Please set the organism attribute to one of %s" %
                self._organism_prefixes.keys())

        if BioCarta._all_pathways is None:
            BioCarta._all_pathways = readXML(self._all_pathways_url)

        if self._pathways is None:

            url_pattern = re.compile("http://cgap.nci.nih.gov/Pathways/BioCarta/%s_(.+)[Pp]athway" \
                % (self._organism_prefix))
            is_pathway_url = lambda tag: tag.name == "a" and not tag.has_attr("class")
            self._pathways = BioCarta._all_pathways.findAll(is_pathway_url, href=url_pattern)

            # Now let us select only the name.
            self._pathways = sorted([entry.attrs['href'].rsplit("/", 1)[1]
                              for entry in self._pathways])
        return self._pathways
    all_pathways = property(_get_all_pathways)

    def get_pathway_protein_names(self, pathway):
        """returns list of genes for the corresponding pathway

        This function scans an HTML page. We have not found another way to 
        get the gene list in a more reobust way. This function was tested on 
        one pathway. Please use with caution.


        """
        self.services.logging.info("Fetching the pathway")
        # first identify gene from GeneInfo tag
        # this is not XML but HTML
        url = "http://cgap.nci.nih.gov/Pathways/BioCarta/%s" % pathway
        html_doc = urlopen(url).read()
        soup = BeautifulSoup(html_doc, 'html.parser')
        links = soup.find_all('area')
        links = [link for link in links if 'GeneInfo' in link.get('href')]

        links = set([link.attrs['href'] for link in links])

        self.services.logging.info("Scanning information about %s genes" % len(links))
        # open each page and get info
        genes = {}
        for link in links:
            html_doc = urlopen(link).read()
            soup = BeautifulSoup(html_doc, 'html.parser')

            table_gene_info = soup.findAll("table")[1]

            gene_name = link.rsplit("=", 1)[1]
            self.services.logging.info(" - " + gene_name)

            genes[gene_name] = {}
            self.tt = table_gene_info
            for row in table_gene_info.find_all('tr'):
                entry = row.find_all('td')
                try:
                    key = entry[0].text.strip()
                except:
                    continue
                try:
                    value = entry[1].text.strip()
                except:
                    continue
                if "[Text]" in key:
                    continue
                genes[gene_name][key] = value

        return genes



#
#  This file is part of bioservices software
#
#  Copyright (c) 2013-2014 - EMBL-EBI
#
#  File author(s): 
#      
#
#  Distributed under the GPLv3 License.
#  See accompanying file LICENSE.txt or copy at
#      http://www.gnu.org/licenses/gpl-3.0.html
#
#  website: https://github.com/cokelaer/bioservices
#  documentation: http://packages.python.org/bioservices
#
##############################################################################
"""Interface to the NCBIBLAST web service

.. topic:: What is NCBIBLAST ?

    :URL: http://blast.ncbi.nlm.nih.gov/
    :service: http://www.ebi.ac.uk/Tools/webservices/services/sss/ncbi_blast_rest

    .. highlights::

        "NCBI BLAST - Protein Database Query

        The emphasis of this tool is to find regions of sequence similarity,
        which will yield functional and evolutionary clues about the structure
        and function of your novel sequence."

        -- from NCBIblast web page


"""
import sys
import time

from bioservices.services import REST
from bioservices import logger
logger.name = __name__


__all__ = ["NCBIblast"]


class NCBIblast():
    """Interface to the `NCBIblast <http://blast.ncbi.nlm.nih.gov/>`_ service.

    ::

        >>> from bioservices import *
        >>> s = NCBIblast(verbose=False)
        >>> jobid = s.run(program="blastp", sequence=s._sequence_example,
            stype="protein", database="uniprotkb", email="name@provider")
        >>> s.getResult(jobid, "out")

    .. warning:: It is very important to provide a real e-mail address as your
        job otherwise very likely will be killed and your IP, Organisation or
        entire domain black-listed.

    When running a blast request, a program is required. You can obtain the
    list using::

        >>> s.parametersDetails("program")
        [u'blastp', u'blastx', u'blastn', u'tblastx', u'tblastn']

    * blastn: Search a nucleotide database using a nucleotide query
    * blastp: Search protein database using a protein query
    * blastx: Search protein database using a translated nucleotide query
    * tblastn     Search translated nucleotide database using a protein query
    * tblastx     Search translated nucleotide database using a translated nucleotide query

    """

    _sequence_example = "MDSTNVRSGMKSRKKKPKTTVIDDDDDCMTCSACQSKLVKISDITKVSLDYINTMRGNTLACAACGSSLKLLNDFAS"
    def __init__(self, verbose=False):
        """.. rubric:: NCBIblast constructor

        :param bool verbose: prints informative messages

        """
        url = "http://www.ebi.ac.uk/Tools/services/rest/ncbiblast"
        self.services = REST(name="NCBIblast", url=url, verbose=verbose)
        self._parameters = None
        self._parametersDetails = {}
        self.checkInterval = 2

    def get_parameters(self):
        """List parameter names.

        :returns: An XML document containing a list of parameter names.

        ::

            >>> from bioservices import ncbiblast 
            >>> n = ncbiblast.NCBIblast()
            >>> res = n.get_parameters()
            >>> [x.text for x in res.findAll("id")]

        .. seealso:: :attr:`parameters` to get a list of the parameters without
            need to process the XML output.
        """

        res = self.services.http_get("parameters", frmt="json", 
                 headers={
                    "User-Agent": self.services.getUserAgent(), 
                    "Accept": "application/json"})
        return res['parameters']

    def _get_parameters(self):
        if self._parameters:
            return self._parameters
        else:
            # on 2 lines in case it fails, self._parameters remaisn None
            res = self.get_parameters()
            self._parameters = res
        return self._parameters
    parameters = property(_get_parameters)

    def get_parameter_details(self, parameterId):
        """Get detailed information about a parameter.

        :returns: An XML document providing details about the parameter or a list
            of values that can take the parameters if the XML could be parsed.

        For example::

            >>> s.parameter_details("matrix")
            [u'BLOSUM45',
             u'BLOSUM50',
             u'BLOSUM62',
             u'BLOSUM80',
             u'BLOSUM90',
             u'PAM30',
             u'PAM70',
             u'PAM250']

        """
        if parameterId not in self.parameters:
            raise ValueError("Invalid parameterId provided(%s). See parameters attribute" % parameterId)

        if parameterId not in self._parametersDetails.keys():
            request = "parameterdetails/" + parameterId
            res = self.services.http_get(request, frmt="json",
                 headers={
                    "User-Agent": self.services.getUserAgent(), 
                    "Accept": "application/json"})

            try:
                data = [x['value'] for x in res["values"]["values"]]
            except:
                data = res
            self._parametersDetails[parameterId] = data
        return self._parametersDetails[parameterId]

    def run(self, program=None, database=None, sequence=None,stype="protein", email=None, **kargs):
        """ Submit a job with the specified parameters.

        .. python ncbiblast_urllib2.py -D ENSEMBL --email "test@yahoo.com" --sequence
        .. MDSTNVRSGMKSRKKKPKTTVIDDDDDCMTCSACQSKLVKISDITKVSLDYINTMRGNTLACAACGSSLKLLNDFAS
        .. --program blastp --database uniprotkb


        .. rubric:: Compulsary arguments

        :param str program: BLAST program to use to perform the search (e.g., blastp)
        :param str sequence: query sequence. The use of fasta formatted sequence is recommended.
        :param list database: list of database names for search or possible a single string (for one database).
            There are some mismatch between the output of parametersDetails("database") and
            the accepted values. For instance UniProt Knowledgebase should be
            given as "uniprotkb".
        :param str email: a valid email address. Will be checked by the service itself.

        .. rubric:: Optional arguments. If not provided, a default value will be used

        :param str type: query sequence type in 'dna', 'rna' or 'protein' (default is protein).
        :param str matrix: scoring matrix to be used in the search (e.g., BLOSUM45).
        :param bool gapalign:  perform gapped alignments.
        :param int alignments:     maximum number of alignments displayed in the output.
        :param exp:     E-value threshold.
        :param bool filter:  low complexity sequence filter to process the query
            sequence before performing the search.
        :param int scores:     maximum number of scores displayed in the output.
        :param int dropoff:     amount score must drop before extension of hits is halted.
        :param match_scores:     match/miss-match scores to generate a scoring matrix 
            for nucleotide searches.
        :param int gapopen:     penalty for the initiation of a gap.
        :param int gapext:     penalty for each base/residue in a gap.
        :param seqrange: region of the query sequence to use for the search. 
            Default: whole sequence.
        :return: A jobid that can be analysed with :meth:`getResult`,
            :meth:`getStatus`, ...

        The up to data values accepted for each of these parameters can be
        retrieved from the :meth:`get_parameter_details`.

        For instance,::

            from bioservices import NCBIblast
            n = NCBIblast()
            n.get_parameter_details("program")

        Example::

            jobid = n.run(program="blastp",
                 sequence=n._sequence_example,
                 stype="protein",
                 database="uniprotkb",
                 email="test@yahoo.fr")

        Database can be a list of databases::

            database=["uniprotkb", "uniprotkb_swissprot"]

        The returned object is a jobid, which status can be checked. It must be
        finished before analysing/geeting the results.

        .. seealso:: :meth:`getResult`

        .. warning:: Cases are not important. Spaces in the database case should 
            be replaced by underscore.

        .. note:: database returned by the server have meaningless names since
            they do not map to the expected names. An example is "ENA Sequence Release" 
            that should be provided as em_rel

        http://www.ebi.ac.uk/Tools/sss/ncbiblast/help/index-nucleotide.html

        """
        # There are compulsary arguments:
        if program is None or sequence is None or database is None or email is None:
            raise ValueError("program, sequence, email  and database must be provided")

        checkParam = self.services.devtools.check_param_in_list

        # Here, we will check the arguments values (not the type)
        # Arguments will be checked by the service itself but if we can
        # catch some before, it is better
        checkParam(program, self.get_parameter_details("program"))
        checkParam(stype, ["protein", "dna", "rna"])

        # So far, we have these parameters
        params = {
            'program': program,
            'sequence': sequence,
            'email': email,
            'stype': stype}

        # all others are optional (actually type is also optional)
        # We can check all of the optional argument provided automatically.
        # this is fine for now but note for instance that stype could not be put
        # here because what is returned by parametersDetails is not exactly what
        # is expected.
        for k, v in kargs.items():
            #print(k, v)
            checkParam(v,self.get_parameter_details(k))
            params[k] = v

        # similarly for the database, we must process it by hand because ther
        # can be more than one database
        #checkParam(database.lower(), [str(x.replace(" ", "_").lower())
        #    for x in self.parametersDetails("database")])
        if isinstance(database, list):
            databases = database[:]
        elif isinstance(database, str):
            databases = [database]
        else:
            raise TypeError("database must be a string or a list of strings")
        params['database'] = databases

        """
parser.add_option('--seqrange', help='region within input to use as query')
# General options
parser.add_option('--title', help='job title')
parser.add_option('--outfile', help='file name for results')
parser.add_option('--outformat', help='output format for results')
parser.add_option('--async', action='store_true', help='asynchronous mode')
parser.add_option('--jobid', help='job identifier')
parser.add_option('--polljob', action="store_true", help='get job result')
parser.add_option('--status', action="store_true", help='get job status')
parser.add_option('--resultTypes', action='store_true', help='get result types')
    """
        # IMPORTANT: use data parameter, not params !!!
        res = self.services.http_post("run", frmt=None, data=params,
            headers={
                 "User-Agent": self.services.getUserAgent(), 
                "accept": "text/plain"})

        return res

    def get_status(self, jobid):
        """Get status of a submitted job

        :param str jobid:
        :param str jobid: a job identifier returned by :meth:`run`.
        :return: A string giving the jobid status (e.g. FINISHED).

         The values for the status are:

         *   RUNNING: the job is currently being processed.
         *   FINISHED: job has finished, and the results can then be retrieved.
         *   ERROR: an error occurred attempting to get the job status.
         *   FAILURE: the job failed.
         *   NOT_FOUND: the job cannot be found.


        """
        res = self.services.http_get("status/{}".format(jobid), frmt="txt", 
            headers={
                "User-Agent": self.services.getUserAgent(), 
                "accept": "text/plain"})
        return res

    def get_result_types(self, jobid):
        """ Get available result types for a finished job.

        :param str jobid: a job identifier returned by :meth:`run`.
        :param bool verbose: print the identifiers together with their label,
            mediaTypes, description and filesuffix.

        :return: A dictionary, which keys correspond to the identifiers. Each
            identifier is itself a dictionary containing the label, description,
            file suffix and mediaType of the identifier.
        """
        if self.get_status(jobid) != 'FINISHED':
            self.services.logging.warning("waiting for the job to be finished. May take a while")
            self.wait(jobid, verbose=False)
        url = 'resulttypes/' + jobid
        res = self.services.http_get(url, frmt="json", 
            headers={
                "User-Agent": self.services.getUserAgent(), 
                "accept": "application/json"})
        return [x["identifier"] for x in res['types']]

    def get_result(self, jobid, result_type):
        """ Get the job result of the specified type.


        :param str jobid: a job identifier returned by :meth:`run`.
        :param str  result_type: type of result to retrieve. See :meth:`getResultTypes`.

        The output from the tool itself.
        Use the 'format' parameter to retireve the output in different formats,
        the 'compressed' parameter to retrieve the xml output in compressed form.
        Format options::

           0 = pairwise,
           1 = query-anchored showing identities,
           2 = query-anchored no identities,
           3 = flat query-anchored showing identities,
           4 = flat query-anchored no identities,
           5 = XML Blast output,
           6 = tabular,
           7 = tabular with comment lines,
           8 = Text ASN.1,
           9 = Binary ASN.1,
           10 = Comma-separated values,
           11 = BLAST archive format (ASN.1).

      See NCBI Blast documentation for details.
      Use the 'compressed' parameter to return the XML output in compressed form.
      e.g. '?format=5&compressed=true'.


        """
        if self.get_status(jobid)!='FINISHED':
            self.services.logging.warning("waiting for the job to be finished. May take a while")
            self.wait(jobid)
        if self.get_status(jobid) != "FINISHED":
            raise ValueError("job is not finished")
        url = 'result/' + jobid + '/' + result_type

        if result_type in ['out', "error", "sequence", "ids"]:
            res = self.services.http_get(url, frmt="txt",
                 headers={
                     "User-Agent": self.services.getUserAgent(),
                     "accept": "text/plain"})
        elif result_type in ['xml']:
            res = self.services.http_get(url, frmt="xml",
                 headers={
                     "User-Agent": self.services.getUserAgent(),
                     "accept": "text/plain"})
        return res

    def wait(self, jobId):
        """This function checks the status of a jobid while it is running

        :param str jobid: a job identifier returned by :meth:`run`.
        :param int checkInterval: interval between requests in seconds.

        """

        if self.checkInterval<1:
            raise ValueError("checkInterval must be positive and less than a second")
        result = 'PENDING'
        while result == 'RUNNING' or result == 'PENDING':
            result = self.get_status(jobId)
            if result == 'RUNNING' or result == 'PENDING':
                time.sleep(self.checkInterval)
        return result

    def _get_database(self):
        return self.get_parameter_details("database")
    databases = property(_get_database,
        doc=r"""Returns accepted databases.""")




#
#  This file is part of bioservices software
#
#  Copyright (c) 2013-2020 - EBI-EMBL - Institut Pasteur
#
#  File author(s):
#      Thomas Cokelaer <cokelaer@ebi.ac.uk>
#      Thomas Cokelaer <thomas.cokelaer@pasteur.fr>
#
#
#  Distributed under the GPLv3 License.
#  See accompanying file LICENSE.txt or copy at
#      http://www.gnu.org/licenses/gpl-3.0.html
#
#  website: https://github.com/cokelaer/bioservices
#  documentation: http://bioservices.readthedocs.io
#
##############################################################################
"""Interface to the mygeneinfo web Service.

.. topic:: What is MyGeneInfo ?

    :URL: https://mygene.info
    :REST: https://mygeneinfo/v3.api/

    .. highlights::

        MyGene.info provides simple-to-use REST web services to query/retrieve gene
        annotation data. It’s designed with simplicity and performance emphasized. You
        can use it to power a web application which requires querying genes and
        obtaining common gene annotations. For example, MyGene.info services are used to
        power BioGPS; or use it in an analysis pipeline to retrieve always up-to-date
        gene annotations.

        -- mygene.info home page, June 2020

"""

from bioservices.services import REST

__all__ = ["MyGeneInfo"]


class MyGeneInfo():
    """Interface to `mygene.infoe <http://mygene.info>`_ service

    .. doctest::

        >>> from bioservices import MyGeneInfo
        >>> s = MyGeneInfoe()

    """

    def __init__(self, verbose=False, cache=False):
        """.. rubric:: Constructor

        :param bool verbose: prints informative messages (default is off)

        """
        url = "https://mygene.info/v3"
        self.services = REST(name="PDBe", url=url, verbose=verbose, cache=cache)

    def get_genes(self, ids, fields="symbol,name,taxid,entrezgene,ensemblgene", 
            species=None, dotfield=True, email=None):
        """Get matching gene objects for a list of gene ids


        :param ids: list of geneinfo IDs
        :param str fields: a comma-separated fields to limit the fields returned
            from the matching gene hits. The supported field names can be found from any
            gene object (e.g. http://mygene.info/v3/gene/1017). Note that it supports dot
            notation as well, e.g., you can pass "refseq.rna". If "fields=all", all
            available fields will be returned. Default:
            "symbol,name,taxid,entrezgene,ensemblgene".
        :param str species:  can be used to limit the gene hits from given
            species. You can use "common names" for nine common species (human, mouse, rat,
            fruitfly, nematode, zebrafish, thale-cress, frog and pig). All other species,
            you can provide their taxonomy ids. Multiple species can be passed using comma
            as a separator. Default: human,mouse,rat.
        :param dotfield: control the format of the returned fields when passed
            "fields" parameter contains dot notation, e.g. "fields=refseq.rna". If True
            the returned data object contains a single "refseq.rna" field, otherwise
            (False), a single "refseq" field with a sub-field of "rna". Default:
            True.
        :param str email": If you are regular users of this services, the
            mygeneinfo maintainers/authors encourage you to provide an email, 
            so that we can better track the usage or follow up with you.

        ::

            mgi = MyGeneInfoe()
            mgi.get_genes(("301345,22637"))
            # first one is rat, second is mouse. This will return a 'notfound'
            # entry and the second entry as expected.
            mgi.get_genes("301345,22637", species="mouse") 

        """
        params = {"ids": ids, "fields":fields}
        if email:                        # pragma: no cover
            params["email"] = email

        assert dotfield in [True, False]
        params["dotfield"] = dotfield

        if species:
            params["species"] = species

        res = self.services.http_post("gene", #params=params, 
            data=params, frmt="json",headers={
            "User-Agent": self.services.getUserAgent(),
            "accept": "application/json",
            "Content-Type": "application/x-www-form-urlencoded"}
        )
        return res

    def get_one_gene(self, geneid, fields="symbol,name,taxid,entrezgene,ensemblgene", 
        dotfield=True, email=None):
        """Get matching gene objects for one gene id

        :param geneid: a valid gene ID
        :param str fields: a comma-separated fields to limit the fields returned
            from the matching gene hits. The supported field names can be found from any
            gene object (e.g. http://mygene.info/v3/gene/1017). Note that it supports dot
            notation as well, e.g., you can pass "refseq.rna". If "fields=all", all
            available fields will be returned. Default:
            "symbol,name,taxid,entrezgene,ensemblgene".
        :param dotfield: control the format of the returned fields when passed
            "fields" parameter contains dot notation, e.g. "fields=refseq.rna". If True
            the returned data object contains a single "refseq.rna" field, otherwise
            (False), a single "refseq" field with a sub-field of "rna". Default:
            True.
        :param str email": If you are regular users of this services, the
            mygeneinfo maintainers/authors encourage you to provide an email, 
            so that we can better track the usage or follow up with you.

        ::

            mgi = MyGeneInfoe()
            mgi.get_genes("301345")
        """
        params = {"ids": geneid, "fields": fields}
        if email:                        # pragma: no cover
            params["email"] = email

        assert dotfield in [True, False]
        params["dotfield"] = dotfield

        res = self.services.http_get(f"gene/{geneid}", params=params, frmt="json")
        return res

    def get_one_query(self, query, email=None, dotfield=True,
            fields="symbol,name,taxid,entrezgene,ensemblgene",
            species="human,mouse,rat", size=10,  _from=0, sort=None,
            facets=None, entrezonly=False, ensemblonly=False):
        """Make gene query and return matching gene list. Support JSONP and CORS as well.

        :param str query: Query string. Examples "CDK2", "NM_052827", "204639_at",
            "chr1:151,073,054-151,383,976", "hg19.chr1:151073054-151383976". The detailed
            query syntax can be found from our docs.
        :param str fields: a comma-separated fields to limit the fields returned
            from the matching gene hits. The supported field names can be found from any
            gene object (e.g. http://mygene.info/v3/gene/1017). Note that it supports dot
            notation as well, e.g., you can pass "refseq.rna". If "fields=all", all
            available fields will be returned. Default:
            "symbol,name,taxid,entrezgene,ensemblgene".
        :param str species: can be used to limit the gene hits from given species. You can use
            "common names" for nine common species (human, mouse, rat, fruitfly, nematode,
            zebrafish, thale-cress, frog and pig). All other species, you can provide their
            taxonomy ids. Multiple species can be passed using comma as a separator.
            Default: human,mouse,rat.
        :param int size: the maximum number of matching gene hits to return
            (with a cap of 1000 at the moment). Default: 10.
        :param int _from: the number of matching gene hits to skip, starting
            from 0. Combining with "size" parameter, this can be useful for paging. Default:
            0.      
        :param sort: the comma-separated fields to sort on. Prefix with "-" for
            descending order, otherwise in ascending order. Default: sort by matching scores
            in decending order.
        :param str facets: a single field or comma-separated fields to return
            facets, for example, "facets=taxid", "facets=taxid,type_of_gene".
        :param bool entrezonly: when passed as True, the query returns only the hits 
            with valid Entrez gene ids. Default: False.
        :param bool ensembleonly: when passed as True, the query returns only the hits 
            with valid Ensembl gene ids. Default: False.
        :param dotfield: control the format of the returned fields when passed
            "fields" parameter contains dot notation, e.g. "fields=refseq.rna". If True
            the returned data object contains a single "refseq.rna" field, otherwise
            (False), a single "refseq" field with a sub-field of "rna". Default:
            True.
        :param str email": If you are regular users of this services, the
            mygeneinfo maintainers/authors encourage you to provide an email, 
            so that we can better track the usage or follow up with you.




        """
        params = {"fields": fields, "size":size, "from":_from}
        if email:   # pragma: no cover
            params["email"] = email

        assert dotfield in [True, False]
        params["dotfield"] = dotfield

        if sort:    
            params["sort"] = sort
        if facets:   # pragma: no cover
            params["facets"] = sort
        assert entrezonly in [True, False]
        params["entrezonly"] = entrezonly
        assert ensemblonly in [True, False]
        params["ensemblonly"] = entrezonly

        res = self.services.http_get(f"query?q={query}", params=params, frmt="json")
        return res

    def get_queries(self, query, email=None, dotfield=True, scopes="all",
            species="human,mouse,rat",
            fields="symbol,name,taxid,entrezgene,ensemblgene", 
            ):
        """Make gene query and return matching gene list. Support JSONP and CORS as well.

        :param str query: Query string. Examples "CDK2", "NM_052827", "204639_at",
            "chr1:151,073,054-151,383,976", "hg19.chr1:151073054-151383976". The detailed
            query syntax can be found from our docs.
        :param str fields: a comma-separated fields to limit the fields returned
            from the matching gene hits. The supported field names can be found from any
            gene object (e.g. http://mygene.info/v3/gene/1017). Note that it supports dot
            notation as well, e.g., you can pass "refseq.rna". If "fields=all", all
            available fields will be returned. Default:
            "symbol,name,taxid,entrezgene,ensemblgene".
        :param str species: can be used to limit the gene hits from given species. You can use
            "common names" for nine common species (human, mouse, rat, fruitfly, nematode,
            zebrafish, thale-cress, frog and pig). All other species, you can provide their
            taxonomy ids. Multiple species can be passed using comma as a separator.
            Default: human,mouse,rat.
        :param dotfield: control the format of the returned fields when passed
             "fields" parameter contains dot notation, e.g. "fields=refseq.rna". If True
             the returned data object contains a single "refseq.rna" field, otherwise
             (False), a single "refseq" field with a sub-field of "rna". Default:
             True.
        :param str email": If you are regular users of this services, the
            mygeneinfo maintainers/authors encourage you to provide an email, 
            so that we can better track the usage or follow up with you.
        :param str scopes: not documented. Set to 'all'

        """
        params = {"q": query, "fields": fields, "scopes": scopes}
        if email:  # pragma: no cover
            params["email"] = email
        assert dotfield in [True, False]
        params["dotfield"] = dotfield

        res = self.services.http_post("query", 
            params=params, frmt="json",headers={
            "User-Agent": self.services.getUserAgent(),
            "accept": "application/json",
            "Content-Type": "application/x-www-form-urlencoded"})
        return res

    def get_metadata(self):
        res = self.services.http_get(f"metadata", frmt="json")
        return res
    def get_taxonomy(self):
        res = self.services.http_get(f"metadata", frmt="json")
        return res['taxonomy']










#!/usr/bin/python
# -*- coding: latin-1 -*-
#
#  This file is part of bioservices software
#
#  Copyright (c) 2013-2014 - EBI-EMBL
#
#  File author(s):
#
#
#  Distributed under the GPLv3 License.
#  See accompanying file LICENSE.txt or copy at
#      http://www.gnu.org/licenses/gpl-3.0.html
#
#  website: https://github.com/cokelaer/bioservices
#  documentation: http://packages.python.org/bioservices
#
##############################################################################
#$Id$
"""This module provides a class :class:`~BioModels` to access to BioModels WS.


.. topic:: What is BioModels ?

    :URL: http://www.ebi.ac.uk/biomodels/
    :Service: http://www.ebi.ac.uk/biomodels
    :Citations: please visit https://www.ebi.ac.uk/biomodels/citation for details

    .. highlights::

        "BioModels is a repository of mathematical models of biological and biomedical
        systems. It hosts a vast selection of existing literature-based physiologically
        and pharmaceutically relevant mechanistic models in standard formats. Our
        mission is to provide the systems modelling community with reproducible,
        high-quality, freely-accessible models published in the scientific literature."

        -- From BioModels website, March 2020


"""
import os
import copy
import webbrowser
from functools import wraps
from bioservices import logger
logger.name = __name__

from bioservices.services import REST

try:
    #python 3
    from urllib.request import urlopen
except:
    from urllib2  import urlopen

__all__ = ["BioModels"]



class BioModels(REST):
    """Interface to the `BioModels <http://www.ebi.ac.uk/biomodels>`_ service

    ::

        from bioservices import BioModels
        bm = BioModels()
        model = bm.get_model('BIOMD0000000299')


    Previous API had several functions such as *getAuthorsByModelId*. This is
    easy to mimic with the new API::

        bm = BioModels()
        models = bm.get_all_models()
        [x['submitter'] for x in res if x[] == "MODEL1204280003"][0]

    This is also true for *getDateLastModifByModelId* and *getModelNameById* if
    one use the field *lastModified* or *name*. There was the ability to 
    search for models based on their CHEBI identifiers, which is not 
    supported anymore; this concerns functions
    *getModelsIdByChEBI*, *getModelsIdByChEBIId*, *getSimpleModelsByChEBIIds*,
    *getSimpleModelsRelatedWithChEBI*. For other searches related to Reactome,
    Uniprot identifiers or GO terms, the :meth:`search` method should work::

        bm.search("P10113")
        bm.search("REACT_33")
        bm.search("GO:0006919")

    """
    _url = "https://www.ebi.ac.uk/biomodels"
    _id_example = "BIOMD0000000100"

    def __init__(self, verbose=True):
        """.. rubric:: constructor

        :param bool verbose:


        """
        super(BioModels, self).__init__(name="BioModels", url=BioModels._url, 
                                        verbose=verbose)

    def _check_format(self, frmt, supported=['json', 'xml', 'html']):
        if frmt not in supported:
            raise ValueError("Supported format for this function are {}. You provided {}".format(
                supported, frmt))

    def get_all_models(self, chunk=100):
        """Return all models"""
        models = []
        offset = 0
        res = self.search("*.*", numResults=chunk)
        while type(res) is dict and 'models' in res:
            models.extend(res['models'])
            offset += chunk
            res = self.search("*.*", offset=offset, numResults=chunk)
        return models

    def get_model(self, model_id, frmt="json"):
        """Fetch information about a given model at a particular revision."""
        self._check_format(frmt)
        res = self.http_get(model_id, frmt=frmt, params={'format': frmt})
        return res

    def get_model_files(self, model_id, frmt="json"):
        """Extract metadata information of model files of a particular model

        :param model_id: a valid BioModels identifier
        :param frmt: format of the output (json, xml)
        """ 
        self._check_format(frmt, ["xml", "json"])
        res = self.http_get("model/files/{}".format(model_id), frmt=frmt, 
            params={'format': frmt})
        return res

    def get_model_download(self, model_id, filename=None,
        output_filename=None):
        """Download a particular file associated with a given model or all its
        files as a COMBINE archive.

        :param model_id: a valid BioModels identifier
        :param str filename: this is the requested filename to be found in the
            model
        :param str output_filename: if you request a different output filename,
            use this parameter
        :param frmt: format of the output (json, xml, html)
        :return:  nothing. This function save the model into a ZIP file called
            after the model identifier. If parameter *filename* is specified, 
            then the output file is the requested filename (if found)

        ::

            bm.get_model_download("BIOMD0000000100", filename="BIOMD0000000100.png")
            bm.get_model_download("BIOMD0000000100")


        This function can retrieve all files in a ZIP archive or a single image.
        In the example below, we retrieve the PNG and plot it using matplotlib.
        Using your favorite image viewver, you should get a better resolution.
        Or just download the SVG version of the model.

        .. plot::
            :include-source:

            from bioservices import BioModels
            bm = BioModels()
            from easydev import TempFile
            with TempFile(suffix=".png") as fout:
                bm.get_model_download("BIOMD0000000100",
                        filename="BIOMD0000000100.png",
                        output_filename=fout.name)
                from pylab import imshow, imread
                imshow(imread(fout.name), aspect="auto")
        """
        params = {}
        if filename:
            params["filename"] = filename

        res = self.http_get("model/download/{}".format(model_id),  
            params=params)

        if filename:
            self.logging.info("Saving {}".format(filename))
            if output_filename is None:
                output_filename = filename
            with open(output_filename, "wb") as fout:
                fout.write(res.content)
        else:
            self.logging.info("Saving file {}.zip".format(model_id) )
            if output_filename is None:
                output_filename = "{}.zip".format(model_id)
            with open(output_filename, "wb") as fout:
                fout.write(res.content)

    def search(self, query, offset=None, numResults=None, sort=None, frmt="json"):
        """Search models of interest via keywords. 

        Examples: PUBMED:"27869123" to search models associated with the PubMed 
        record identified by 27869123. 

        :param str query: search query. colon character must be escaped
        :param int offset: number of items to skip before starting to collect the
            result set
        :param int numResults: number of items to return
        :param str sort: sort criteria in {id-asc, relevance-asc, relevance-desc,
            first_author-asc, first_author, name-asc, name-desc,
            publication_year-asc, publication_year-desc}
        :param str frmt: format of the output (json, xml)

        """
        self._check_format(frmt, ["xml", "json", "html"])
        params = {"query": query, "format": frmt}
        if offset:
            params["offset"] = offset
        if numResults:
            params["numResults"] = numResults
        if sort:
            params["sort"] = sort

        sort_options = ["id-asc", "relevance-asc", "relevance-desc",
            "first_author-asc", "first_author", "name-asc", "name-desc",
            "publication_year-asc", "publication_year-desc"]
        if sort and sort not in sort_options:
            raise ValueError("sort must be in {}. You provided {}".format(
                sort_options, sort))
        res = self.http_get("search", params=params)
        return res

    def search_download(self, models, output_filename="models.zip", force=False):
        """Returns models (XML) corresponding to a list of model identifiers. 

        :param str models: list of model identifiers using comma to separate
            them. Could be a list of string (e.g 'BIOMD1,BIOMD2' or ['BIOMD1',
            'BIOMD2']
        :param str output_filename: file used to save the models. This is a
            zipped file. If the file exists, you must use the *force** parameter

        .. todo:: if no models are found (typos), an error message is printed.
            if one model is not found, there is no warning or errors. Could be
            nice to have a warning by introspecting the number of models in the
            output file
        """
        if isinstance(models, list):
            models = ",".join(models)
        res = self.http_get("search/download", params={"models": models})

        if res == 404:
            self.logging.error("One of your model ID was probably incorrect")
            return

        self.logging.info(output_filename)
        if os.path.exists(output_filename) and force is False:
            raise IOError("{} exists already. Set force to True or change the output_filename argument".format(output_filename))

        with open(output_filename, "wb") as fout:
            fout.write(res.content)

    def search_parameter(self, query, start=0, size=10, sort=None, frmt="json"):
        """Search for parameters of a model

        **Details** BioModels Parameters is a resource that facilitates easy
        search and retrieval of parameter values used in the SBML models stored in the
        BioModels repository. Users can search for a model entity (e.g. a protein or
        drug) to retrieve the rate equations describing it; the associated parameter
        values and the initial concentration from the SBML models in BioModels. Although
        these data are directly extracted from the curated SBML models, they are not
        individually curated or validated; rather presented as such in the table below.
        Hence BioModels Parameters table will only provide a quick overview of available
        parameter values for guidance and original model should be referred to
        understand the complete context of the parameter usage.


        :param str query: A query to search against the model parameter values.
        :param int start: if is the offset of the result set (default 0)
        :param int size: number of items to display per page
        :param str sort: model or entity
        :param str frmt: the format of the result (xml, csv, json)

        ::

            bm.search_parameter("MAPK", size=100, sort="entity")

        """
        self._check_format(frmt, ["xml", "json", "csv"])
        params = {"format": frmt,
                    "size": size,
                    "start": start,
                    "query":query}
        sort_options = ["model", "entity", None]
        if sort not in sort_options:
            raise ValueError("sort must be in {}. You provided {}".format(
                sort_options, sort))
        if sort:
            params["sort"] = sort

        res = self.http_get("parameterSearch/search", params=params)

        return res

    def get_p2m_missing(self, frmt="json"):
        """Retrieve all models in Path2Models that are now only available indirectly, 
        through the representative model for the corresponding genus

        :param str frmt: the format of the result (xml, csv, json)
        :return: list of model identifiers

        """
        self._check_format(frmt)
        res = self.http_get("p2m/missing", params={"format": frmt})
        res = res['missing']
        self.logging.info("Found {} missing model".format(len(res)))
        return res

    def get_p2m_representative(self, model, frmt="json"):
        """Retrieve a representative model in Path2Models

        Get the representative model identifier for a given missing model in Path2Models. 
        This endpoint accepts as parameters a mandatory model identifier and an 
        optional response format

        :param str model: The identifier of a model of interest
        :param str frmt: the format of the result (xml, csv, json)

        """
        self._check_format(frmt)
        res = self.http_get("p2m/representative",
            params={"format": frmt, "model": model})
        return res

    def get_p2m_representatives(self, models, frmt="json"):
        """Find the replacement accessions for a set of Path2Models entries

        Get the representative model identifiers of a set of given missing
        models in Path2Models. This end point expects a comma-separated list of model
        identifiers (without any surrounding whitespace) and an optional response
        format. Examples: BMID000000112902, BMID000000009880, BMID000000027397.

        :param str model: The model identifiers separated by commas, or as a
            list.
        :param str frmt: the format of the result (xml, csv, json)

        .. doctest::

            from bioservices import BioModels
            bm = BioModels()
            bm.get_p2m_representatives("BMID000000112902, BMID000000009880, BMID000000027397")


        """
        if isinstance(models, list):
            models = ",".join(models)
        else:
            # Get rid of possible spaces
            models = ",".join([x.strip() for x in models.split(",")])

        self._check_format(frmt)
        res = self.http_get("p2m/representatives", 
            params={"format": frmt, "modelIds": models})
        return res

    def get_pdgsmm_missing(self, frmt="json"):
        """Retrieve the identifiers of all PDGSMM entries that are no longer directly accessible

        :param str frmt: the format of the result (xml, csv, json)
        :return: list of model identifiers
        """
        self._check_format(frmt)
        res = self.http_get("pdgsmm/missing", params={"format": frmt})
        res = res['missing']
        self.logging.info("Found {} missing model".format(len(res)))
        return res

    def get_pdgsmm_representative(self, model, frmt="json"):
        """Retrieve a representative model in PDGSMM

        Get the representative model identifier for a given missing model in
        PDGSMM. This endpoint accepts as parameters a mandatory model identifier and an
        optional response format.

        :param str model: The identifier of a model of interest
        :param str frmt: the format of the result (xml, csv, json)

        """
        self._check_format(frmt)
        res = self.http_get("pdgsmm/representative",
            params={"format": frmt, "model": model})
        return res

    def get_pdgsmm_representatives(self, models, frmt="json"):
        """Find the replacement accessions for a set of PDFSSM

        Get the representative model identifiers of a set of given missing
        models in PDGSMM. This end point expects a comma-separated list of model
        identifiers (without any surrounding whitespace) and an optional response
        format. Examples: MODEL1707110145,MODEL1707112456,MODEL1707115900.

        :param str model: The model identifiers separated by commas, or as a
            list.
        :param str frmt: the format of the result (xml, csv, json)

        """
        if isinstance(models, list):
            models = ",".join(models)
        else:
            # Get rid of possible spaces
            models = ",".join([x.strip() for x in models.split(",")])

        self._check_format(frmt)
        res = self.http_get("pdgsmm/representatives", 
            params={"format": frmt, "modelIds": models})
        return res



#
#  This file is part of bioservices software
#
#  Copyright (c) 2013-2014 - EBI-EMBL
#
#  File author(s):
#      Thomas Cokelaer <cokelaer@ebi.ac.uk>
#
#
#  Distributed under the GPLv3 License.
#  See accompanying file LICENSE.txt or copy at
#      http://www.gnu.org/licenses/gpl-3.0.html
#
#  website: https://github.com/cokelaer/bioservices
#  documentation: http://packages.python.org/bioservices
#
##############################################################################
"""Interface to the ArrayExpress web Service.

.. topic:: What is ArrayExpress ?

    :URL: http://www.ebi.ac.uk/arrayexpress/
    :REST: http://www.ebi.ac.uk/arrayexpress/xml/v2/experiments

    .. highlights::

        ArrayExpress is a database of functional genomics experiments that can be queried and the data downloaded. It includes gene expression data from microarray and high throughput sequencing studies. Data is collected to MIAME and MINSEQE standards. Experiments are submitted directly to ArrayExpress or are imported from the NCBI GEO database.

        -- ArrayExpress home page, Jan 2013

"""

from bioservices.services import REST
from bioservices import logger
logger.name = __name__


__all__ = ["ArrayExpress"]


class ArrayExpress():
    """Interface to the `ArrayExpress <http://www.ebi.ac.uk/arrayexpress>`_ service

    ArrayExpress allows to retrieve data sets used in various experiments.

    **QuickStart** Given an experiment name (e.g., E-MEXP-31), type::

        s = ArrayExpress()
        s.getAE('E-MEXP-31')

    You can also quickyl retrieve experiments matching some search queries as
    follows::

        a.queryAE(keywords="pneumonia", species='homo+sapiens')

    Now let us look at other methods.If you know the file and experiment
    name, you can retrieve a specific file as follows::

        >>> from bioservices import ArrayExpress
        >>> s = ArrayExpress()
        >>> # retrieve a specific file from a experiment
        >>> res = s.retrieveFile("E-MEXP-31", "E-MEXP-31.idf.txt")

    The main issue is that you may not know the experiment you are looking for.
    You can query experiments by keyword::

        >>> # Search for experiments
        >>> res = s.queryExperiments(keywords="cancer+breast", wholewords=True)

    keywords used in queries follows these rules:

    * Accession number and keyword searches are case insensitive
    * More than one keyword can be searched for using the + sign (e.g. keywords="cancer+breast")
    * Use an asterisk as a multiple character wild card (e.g. keywords="colo*")
    * use a question mark ? as a single character wild card (e.g. keywords="te?t")

    More complex queries can be constructed using the operators AND, OR or NOT.
    AND is the default if no operator is specified. Either experiments or
    files can be searched for. Examples are::

        keywords="prostate+AND+breast"
        keywords="prostate+breast"      # same as above
        keywords="prostate+OR+breast"
        keywords="prostate+NOT+breast "

    The returned objects are XML parsed with beautifulSoup. You can get all
    experiments using the getChildren method:

    .. doctest::
        :options: +SKIP

        >>> res = s.queryExperiments(keywords="breast+cancer")
        >>> len(res.getchildren())
        1487


    If you know what you are looking for, you can give the experiment name::

        >>> res = s.retrieveExperiment("E-MEXP-31")
        >>> exp = res.getchildren()[0]   # it contains only one experiment
        >>> [x.text for x in exp.getchildren() if x.tag == "name"]
        ['Transcription profiling of mammalian male germ cells undergoing mitotic
        growth, meiosis and gametogenesis in highly enriched cell populations']

    Using the same example, you can retrieve the names of the files related to
    the experiment::

        >>> files = [x.getchildren() for x in exp.getchildren() if x.tag == "files"]
        >>> [x.get("name") for x in files[0]]
        ['E-MEXP-31.raw.1.zip',
         'E-MEXP-31.processed.1.zip',
         'E-MEXP-31.idf.txt',
         'E-MEXP-31.sdrf.txt']

    New in version 1.3.7 you can use the method :meth:`getEA`

    Then, you may want to download a particular file::

        >>> s.retrieveFile("E-MEXP-31", "E-MEXP-31.idf.txt")


    .. seealso:: :meth:`queryFiles` for more details about the parameters to be
        used in queries.

    .. warning:: supports only new style (v2). You can still use the old style by
        setting the request manually using the :meth:`version`.

    .. warning:: some syntax requires the + character, which is a special character
        for http requests. It is replaced internally by spaces if found
    .. warning:: filtering is not implemented (e.g., assaycount:[x TO y]syntax.)
    """
    def __init__(self, verbose=False, cache=False):
        """.. rubric:: Constructor

        :param bool verbose: prints informative messages

        """
        self.services = REST(name="ArrayExpress",
            url="http://www.ebi.ac.uk/arrayexpress", cache=cache,
            verbose=verbose)

        self.version = "v2"

    def _search(self, mode, **kargs):
        """common function to search for files or experiments"""
        assert mode in ["experiments", "files"]
        url = "{0}/{1}/{2}".format("json", self.version, mode)

        defaults = {
            "accession":None, #ex: E-MEXP-31
            "keywords":None,
            "species": None,
            "wholewords": "on",
            "expdesign":  None,
            "exptype": None,
            "gxa": "true",
            "pmid": None,
            "sa": None,
            "ef": None,         # e.g., CellType
            "efv": None,         # e.g., HeLa
            "array":None,       # ex: A-AFFY-33
            "expandfo": "on",
            "directsub": "true",
            "sortby": ["accession", "name", "assays", "species",
                       "releasedate", "fgem", "raw", "atlas"],
            "sortorder": ["ascending", "descending"],
        }

        for k in kargs.keys():
            if k not in defaults.keys():
                raise ValueError("Incorrect value provided ({}). Correct values are {}".format(
                    k, sorted(defaults.keys())))

        #if len(kargs.keys()):
        #    url += "?"
        params = {}

        for k, v in kargs.items():
            if k in ["expandfo", "wholewords"]:
                if v in ["on", True, "true", "TRUE", "True"]:
                    #params.append(k + "=on")
                    params[k] = "on"
            elif k in ["gxa", "directsub"]:
                if v in ["on", True, "true", "TRUE", "True"]:
                    #params.append(k + "=true")
                    params[k] = "true"
                elif v in [False, "false", "False"]:
                    #params.append(k + "=false")
                    params[k] = "false"
                else:
                    raise ValueError("directsub must be true or false")
            else:
                if k in ["sortby", "sortorder"]:
                    self.services.devtools.check_param_in_list(v, defaults[k])
                #params.append(k + "=" + v)
                params[k] = v

        # NOTE: + is a special character that is replaced by %2B
        # The + character is the proper encoding for a space when quoting
        # GET or POST data. Thus, a literal + character needs to be escaped
        # as well, lest it be decoded to a space on the other end
        for k,v in params.items():
            params[k] = v.replace("+",  " ")

        self.services.logging.info(url)
        res = self.services.http_get(url, frmt="json", params=params)
        return res

    def queryFiles(self, **kargs):
        """Retrieve a list of files associated with a set of experiments

        The following parameters are used to search for experiments/files:

        :param str accession: experiment primary or secondary accession e.g. E-MEXP-31
        :param str array: array design accession or name e.g., A-AFFY-33
        :param str ef: Experimental factor, the name of the main variables in an
            experiment. (e.g., CellType)
        :param str efv:  Experimental factor value. Has EFO expansion. (e.g.,
            HeLa)
        :param str expdesign: Experiment design type  (e.g., "dose+response")
        :param str exptype:  Experiment type. Has EFO expansion. (e.g.,
            "RNA-seq")
        :param str gxa: Presence in the Gene Expression Atlas. Only value is gxa=true.
        :param str keywords: e.g. "cancer+breast"
        :param str pmid: PubMed identifier (e.g., 16553887)
        :param str sa: Sample attribute values. Has EFO expansion. fibroblast
        :param str species: Species of the samples.Has EFO expansion. (e.g., "homo+sapiens")
        :param bool wholewords:

        The following parameters can filter the experiments:

        :param str directsub: only experiments directly submitted to
            ArrayExpress (true) or only imported from GEO databae (false)


        The following parameters can sort the results:

        :param str sortby: sorting by grouping (can be accession, name, assays,
            species, releasedata, fgem, raw, atlas)
        :param str sortorder: sorting by orderering. Can be either ascending or
            descending (default)

        .. doctest::
            :options: +SKIP

            >>> from bioservices import ArrayExpress
            >>> s = ArrayExpress()
            >>> res = s.queryFiles(keywords="cancer+breast", wholewords=True)
            >>> res = s.queryExperiments(array="A-AFFY-33", species="Homo Sapiens")
            >>> res = s.queryExperiments(array="A-AFFY-33", species="Homo Sapiens",
            ...                          sortorder="releasedate")
            >>> res = s.queryExperiments(array="A-AFFY-33", species="Homo+Sapiens",
            ...     expdesign="dose response", sortby="releasedate", sortorder="ascending")
            >>> dates = [x.findall("releasedate")[0].text for x in res.getchildren()]

        """
        res = self._search("files", **kargs)
        return res

    def queryExperiments(self, **kargs):
        """Retrieve experiments

        .. seealso:: :meth:`~bioservices.arrayexpress.ArrayExpress.queryFiles` for
            all possible keywords

        .. doctest::
            :options: +SKIP

            >>> res = s.queryExperiments(keywords="cancer+breast", wholewords=True)

        """
        res = self._search("experiments", **kargs)
        return res

    def retrieveExperiment(self, experiment):
        """alias to queryExperiments if you know the experiment name

        ::

            >>> s.retrieveExperiment("E-MEXP-31")
            >>> # equivalent to
            >>> s.queryExperiments(accession="E-MEXP-31")

        """
        res = self.queryExperiments(keywords=experiment)
        return res

    def retrieveFile(self, experiment, filename, save=False):
        """Retrieve a specific file from an experiment

        :param str filename:

        ::

            >>> s.retrieveFile("E-MEXP-31", "E-MEXP-31.idf.txt")
        """
        files = self.retrieveFilesFromExperiment(experiment)

        assert filename in files, """Error. Provided filename does not seem to be correct.
            Files available for %s experiment are %s """ % (experiment, files)

        url =  "files/" + experiment + "/" + filename

        if save:
            res = self.services.http_get(url, frmt="txt")
            f = open(filename,"w")
            f.write(res)
            f.close()
        else:
            res = self.services.http_get(url, frmt="txt")
            return  res

    def retrieveFilesFromExperiment(self, experiment):
        """Given an experiment, returns the list of files found in its description


        :param str experiment: a valid experiment name
        :return: the experiment files

        .. doctest::

            >>> from bioservices import ArrayExpress
            >>> s = ArrayExpress(verbose=False)
            >>> s.retrieveFilesFromExperiment("E-MEXP-31")
            ['E-MEXP-31.raw.1.zip', 'E-MEXP-31.processed.1.zip', 'E-MEXP-31.idf.txt', 'E-MEXP-31.sdrf.txt']


        """
        res = self.queryExperiments(keywords=experiment)
        exp = res['experiments']['experiment']
        files = exp['files']
        output = [v['name'] for k,v in files.items() if k]
        return output

    def queryAE(self, **kargs):
        """Returns list of experiments

        See :meth:`queryExperiments` for parameters and usage

        This is a wrapper around :meth:`queryExperiments` that returns only
        the accession values.

        ::

            a.queryAE(keywords="pneumonia", species='homo+sapiens')
        """
        sets = self.queryExperiments(**kargs)
        return [x['accession'] for x in sets['experiments']['experiment']]

    def getAE(self, accession, type='full'):
        """retrieve all files from an experiments and save them locally"""
        filenames = self.retrieveFilesFromExperiment(accession)
        self.services.logging.info("Found %s files" % len(filenames))
        for i,filename in enumerate(filenames):
            res = self.retrieveFile(accession, filename)
            if filename.endswith('.zip'):
                with open(filename, 'wb') as fout:
                    self.services.logging.info("Downloading %s" % filename)
                    fout.write(res)
            else:
                with open(filename, 'w') as fout:
                    self.services.logging.info("Downloading %s" % filename)
                    fout.write(res)
                



#
#  This file is part of bioservices software
#
#  Copyright (c) 2013-2014 - EBI-EMBL
#
#  File author(s):
#      
#
#  Distributed under the GPLv3 License.
#  See accompanying file LICENSE.txt or copy at
#      http://www.gnu.org/licenses/gpl-3.0.html
#
#  website: https://github.com/cokelaer/bioservices
#  documentation: http://packages.python.org/bioservices
#
##############################################################################
"""This module provides a class :class:`~BioDBNet` to access to BioDBNet WS.


.. topic:: What is BioDBNet ?

    :URL: http://biodbnet.abcc.ncifcrf.gov/
    :Service: http://biodbnet.abcc.ncifcrf.gov/webServices
    :Citations:  Mudunuri,U., Che,A., Yi,M. and Stephens,R.M. (2009) bioDBnet: the biological database network. Bioinformatics, 25, 555-556

    .. highlights::

        "BioDBNet Database is a repository hosting computational models of biological
        systems. A large number of the provided models are published in the
        peer-reviewed literature and manually curated. This resource allows biologists
        to store, search and retrieve mathematical models. In addition, those models can
        be used to generate sub-models, can be simulated online, and can be converted
        between different representational formats. "

        -- From BioDBNet website, Dec. 2012

    .. versionadded:: 1.2.3
    .. sectionauthor:: Thomas Cokelaer, Feb 2014

"""
import io
from bioservices.services import REST
from bioservices import logger
logger.name = __name__

try:
    import pandas as pd
except:
    pass

__all__ = ["BioDBNet"]



class BioDBNet():
    """Interface to the `BioDBNet <http://biodbnet.abcc.ncifcrf.gov/>`_ service

    ::

        >>> from bioservices import *
        >>> s = BioDBNet()

    Most of the BioDBNet WSDL are available. There are functions added to
    the original interface such as :meth:`extra_getReactomeIds`.

    Use :meth:`db2db` to convert from 1 database to some databases.
    Use :meth:`dbReport` to get the convertion from one database to all
    databases.

    """
    _url = 'https://biodbnet-abcc.ncifcrf.gov/webServices/rest.php/biodbnetRestApi.json'
    def __init__(self, verbose=True, cache=False):
        """.. rubric:: Constructor

        :param bool verbose:

        """
        self.services = REST(name="BioDBNet", url=BioDBNet._url, verbose=verbose, cache=cache)
        self._valid_inputs = self.getInputs() 

    def _list_to_string(self, values):
        if isinstance(values, list):
            values = ",".join(values)
        return values

    def _interpret_output_db(self, input_db, output_db):
        # in biodbnet, the database can be provided as
        # in the output of getInputs() that is with capitals and spaces
        # or with no spaces and no big caps.
        # Here, like in _check_db(), with convert everything to small caps and
        # remove spaces so as to compare the input/output databases with the
        # list of databases returned by getInputs
        outputs = self._list_to_string(output_db)
        inputResult = self.getInputs()
        #getOutputsForInput method
        outputResult = self.getOutputsForInput(input_db)
        outputResult = [this.lower().replace(" ", "") for this in outputResult]
        for output in outputs.split(","):
            if output.lower().replace(" ", "") not in outputResult:
                raise ValueError(output + " not found")
        return outputs

    def _check_db(self, value):
        def convert(value):
            return value.lower().replace(" ", "")
        if convert(value) not in [convert(this) for this in self._valid_inputs]:
            raise ValueError("Invalid value {} not a known database".format(value))

    def db2db(self, input_db, output_db, input_values, taxon=9606):
        """Retrieves models associated to the provided Taxonomy text.

        :param input_db: input database.
        :param output_db: list of databases to map to.
        :param input_values: list of identifiers to map to the output databases
        :return:  dataframe where index correspond to the input database
            identifiers. The columns contains the identifiers for each output
            database (see example here below).

        ::

            >>> from bioservices import BioDBNet
            >>> input_db = 'Ensembl Gene ID'
            >>> output_db = ['Gene Symbol']
            >>> input_values = ['ENSG00000121410', 'ENSG00000171428']
            >>> df = s.db2db(input_db, output_db, input_values, 9606)
                            Gene Symbol
            Ensembl Gene ID
            ENSG00000121410        A1BG
            ENSG00000171428        NAT1

        """
        self._check_db(input_db)
        # This also check that the outputs exist and are compatible with the
        # input.
        outputs = self._interpret_output_db(input_db, output_db)

        url = self._url + "?method=db2db"
        url += "&input={}".format(input_db)
        url += "&outputs={}".format(outputs)
        url += "&inputValues={}".format(self._list_to_string(input_values))
        url += "&taxonId={}".format(taxon)
        url += "&format={}".format("row")
        request = self.services.http_get(url)
        try: # TODO can be removed in v2
            df = pd.DataFrame(request)
            df.set_index("InputValue", inplace=True)
            df.index.name = input_db
            return df
        except Exception as err:
            self.logging.error(err)
            return request

    def dbFind(self, output_db, input_values, taxon="9606"):
        """dbFind method

        dbFind can be used when you do not know the actual type of your identifiers or
        when you have a mixture of different types of identifiers. The tool finds the
        identifier type and converts them into the selected output if the identifiers
        are within the network. 

        :param str output_db: valid database name
        :param list input_values: list of identifiers to look for
        :return: a dataframe with index set to the input values.


        ::

            >>> b.dbFind("Gene ID", ["ZMYM6_HUMAN", "NP_710159", "ENSP00000305919"])
                            Gene ID                Input Type
            InputValue
            ZMYM6_HUMAN        9204        UniProt Entry Name
            NP_710159        203100  RefSeq Protein Accession
            ENSP00000305919  203100        Ensembl Protein ID

        """
        self._check_db(output_db)

        url = self._url + "?method=dbfind"
        url += "&output={}".format(output_db)
        url += "&inputValues={}".format(self._list_to_string(input_values))
        url += "&taxonId={}".format(taxon)
        url += "&format={}".format("row")
        request = self.services.http_get(url)
        try:
            return pd.DataFrame(request).set_index("InputValue")
        except:
            return request

    def dbOrtho(self, input_db, output_db, input_values, input_taxon, output_taxon):
        """Convert identifiers from one species to identifiers of a different species

        :param input_db: input database
        :param output_db: output database
        :param input_values: list of identifiers to retrieve
        :param input_taxon: input taxon
        :param output_taxon: output taxon
        :return:  dataframe where index correspond to the input database
            identifiers. The columns contains the identifiers for each output
            database (see example here below)

        ::

            >>> df = b.dbOrtho("Gene Symbol", "Gene ID", ["MYC", "MTOR", "A1BG"],
            ...                    input_taxon=9606, output_taxon=10090)
                 Gene ID InputValue
            0   17869        MYC
            1   56717       MTOR
            2  117586       A1BG

        """
        self._check_db(input_db)
        self._check_db(output_db)
        url = self._url + "?method=dbortho"
        url += "&input={}".format(input_db)
        url += "&output={}".format(output_db)
        url += "&inputValues={}".format(self._list_to_string(input_values))
        url += "&inputTaxon={}".format(input_taxon)
        url += "&outputTaxon={}".format(output_taxon)
        url += "&format={}".format("row")
        request = self.services.http_get(url)

        try:
            df = pd.DataFrame(request).set_index("InputValue")
            df.index.name = input_db
            return df
        except:
            return request

    def dbReport(self, input_db, input_values, taxon=9606):
        """Same as :meth:`db2db` but returns results for all possible outputs.

        :param input_db: input database
        :param input_values: list of identifiers to retrieve
        :return:  dataframe where index correspond to the input database
            identifiers. The columns contains the identifiers for each output
            database (see example here below)

        ::

            df = s.dbReport("Ensembl Gene ID", ['ENSG00000121410', 'ENSG00000171428'])

        """
        self._check_db(input_db)
        # This also check that the outputs exist and are compatible with the
        # input.
        url = self._url + "?method=dbreport"
        url += "&input={}".format(input_db)
        url += "&inputValues={}".format(self._list_to_string(input_values))
        url += "&taxonId={}".format(taxon)
        url += "&format={}".format("row")
        request = self.services.http_get(url)
        try: # TODO can be removed in v2
            df = pd.DataFrame(request)
            df.set_index("InputValue", inplace=True)
            df.index.name = input_db
            return df
        except Exception as err:
            self.logging.error(err)
            return request
        inputValues = self._interpret_input_db(inputValues)

        # df = pd.readcsv(io.StringIO(res.strip()), sep="\t")

    def dbWalk(self, db_path, input_values, taxon=9606):
        """Walk through biological database network

        dbWalk is a form of database to database conversion where the user has complete
        control on the path to follow while doing the conversion. When a input/node is
        added to the path the input selection gets updated with all the nodes that it
        can access directly.

        :param db_path: path to follow in the databases 
        :param input_values: list of identifiers 
        :return: a dataframe with columns corresponding to the path nodes

        A typical example is to get the Ensembl mouse homologs for  
        Ensembl Gene ID's from human. This conversion is not possible 
        through :meth:`db2db` as Homologene does not have
        Ensembl ID's and the input and output nodes to acheive this would both be
        'Ensembl Gene ID'. It can however be run by using dbWalk as follows.
        Add Ensembl Gene ID to the path, then add Gene Id, Homolog - Mouse Gene ID
        and Ensebml Gene ID to complete the path.

        ::

            db_path = "Ensembl Gene ID->Gene ID->Homolog - Mouse Gene ID->Ensembl Gene ID"
            s.dbWalk(db_path, ["ENSG00000175899"])

        .. todo:: check validity of the path

        """
        url = self._url + "?method=dbwalk"
        url += "&inputValues={}".format(self._list_to_string(input_values))
        url += "&dbPath={}".format(db_path)
        url += "&taxonId={}".format(taxon)
        url += "&format={}".format("row")
        request = self.services.http_get(url)
        try:
            return pd.DataFrame(request)
        except:
            return request

    def getDirectOutputsForInput(self, input_db):
        """Gets all the direct output nodes for a given input node

        Gets all the direct output nodes for a given input node
        Outputs reachable by single edge connection in the bioDBnet graph.

        ::

            b.getDirectOutputsForInput("genesymbol")
            b.getDirectOutputsForInput("Gene Symbol")
            b.getDirectOutputsForInput("pdbid")
            b.getDirectOutputsForInput("PDB ID")
        """
        self._check_db(input_db)

        url = self._url
        url += "?method=getdirectoutputsforinput"
        url += "&input={}&directOutput=1".format(input_db)
        request = self.services.http_get(url)
        try:
            return request['output']
        except:
            return request

    def getInputs(self):
        """Return list of possible input database

        ::

            s.getInputs()
        """
        request = self.services.http_get(self._url + "?method=getinputs")
        try:
            return request['input']
        except:
            return request

    def getOutputsForInput(self, input_db):
        """Return list of possible output database for a given input database

        ::

            s.getOutputsForInput("UniProt Accession")

        """
        self._check_db(input_db)
        url = self._url + "?method=getoutputsforinput"
        url += "&input={}".format(input_db)
        request = self.services.http_get(url)
        try:
            return request['output']
        except:
            return request



from bioservices.apps.fasta import FASTA, MultiFASTA
import tempfile
import pytest

@pytest.mark.flaky
def test_fasta():
    fasta = FASTA()
    fasta.load_fasta(None)
    fasta.load_fasta("P43403")
    fasta.load_fasta("P43403") # already there 
    fasta.header
    fasta.gene_name
    fasta.sequence
    fasta.fasta
    fasta.identifier
    fh = tempfile.NamedTemporaryFile(delete=False)
    fasta.save_fasta(fh.name)
    fasta.read_fasta(fh.name)
    fh.delete = True
    fh.close()

@pytest.mark.flaky
def test_multi_fasta():
    klass = MultiFASTA()

    try:
        klass.fasta
        assert False
    except:
        assert True
    try:
        f.header
        assert False
    except:
        assert True

    klass.load_fasta("P43403")
    klass.load_fasta("P43408")

    assert len(klass) == 2

    klass.ids
    klass.fasta
    fh = tempfile.NamedTemporaryFile(delete=False)
    klass.save_fasta(fh.name)
    klass.read_fasta(fh.name)

    klass.fasta["P43403"]




#
#  This file is part of BioServices
#
#  Distributed under the terms of the 3-clause BSD license.
#  The full license is in the LICENSE file, distributed with this software.
#
#  website: https://github.com/sequana/sequana
#  documentation: http://sequana.readthedocs.io
#
##############################################################################
import functools
import glob
import os
from pathlib import Path
import pkg_resources
import shutil
import subprocess
import sys
import tempfile

import click

import colorlog

logger = colorlog.getLogger(__name__)

__all__ = ["main"]


# This can be used by all commands as a simple decorator
def common_logger(func):
    @click.option(
        "--logger",
        default="INFO",
        type=click.Choice(["INFO", "DEBUG", "WARNING", "CRITICAL", "ERROR"]),
    )
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    return wrapper


CONTEXT_SETTINGS = dict(help_option_names=["-h", "--help"])


from bioservices import version
@click.group(context_settings=CONTEXT_SETTINGS)
@click.version_option(version=version)
def main(**kwargs):
    """This is the main entry point for a set of BioServices application

    """
    pass


@main.command()
@click.option("--accession", type=click.STRING)
@click.option("--output", type=click.STRING, default=None)
@click.option("--method", type=click.Choice(["ENA", "EUtils"]),
    default="EUtils")
def download_accession(**kwargs):
    """

    Input file can be gzipped or not. The --output-file

        bioservices download-accession FN433596.1
    """
    from bioservices.apps.download_fasta import download_fasta
    download_fasta(kwargs['accession'], output_filename=kwargs['output'],
        method=kwargs['method'])




#
#  This file is part of bioservices software
#
#  Copyright (c) 2013-2014 - EMBL-EBI
#
#  File author(s):
#      Sven-Maurice Althoff, Christian Knauth
#      
#
#  Distributed under the GPLv3 License.
#  See accompanying file LICENSE.txt or copy at
#      http://www.gnu.org/licenses/gpl-3.0.html
#
#  website: https://github.com/cokelaer/bioservices
#  documentation: http://packages.python.org/bioservices
#
##############################################################################
"""Interface to the MUSCLE web service

.. topic:: What is MUSCLE ?

    :URL: http://www.drive5.com/muscle/
    :service: http://www.ebi.ac.uk/Tools/webservices/services/msa/muscle_rest

    .. highlights::

        "MUSCLE - (MUltiple Sequence Comparison by Log-Expectation) 1)

        is claimed to achieve both better average accuracy and better speed than
        ClustalW or T-Coffee, depending on the chosen options. Multiple alignments
        of protein sequences are important in many applications, including
        phylogenetic tree estimation, secondary structure prediction and critical
        residue identification."

        -- from EMBL-EBI web page

"""
import sys
import time
from bioservices.services import REST
from bioservices import logger
logger.name = __name__


__all__ = ["MUSCLE"]


class MUSCLE():
    """Interface to the `MUSCLE <http://www.ebi.ac.uk/Tools/webservices/services/msa/muscle_rest>`_ service.

    ::

        >>> from bioservices import *
        >>> m = MUSCLE(verbose=False)
        >>> sequencesFasta = open('filename','r')
        >>> jobid = n.run(frmt="fasta", sequence=sequencesFasta.read(),
                        email="name@provider")
        >>> s.getResult(jobid, "out")

    .. warning:: It is very important to provide a real e-mail address as your
        job otherwise very likely will be killed and your IP, Organisation or
        entire domain black-listed.


    Here is another similar example but we use :class:`~bioservices.uniprot.UniProt`
    class provided in bioservices to fetch the FASTA sequences::


        >>> from bioservices import UniProt, MUSCLE
        >>> u = UniProt(verbose=False)
        >>> f1 = u.get_fasta("P18413")
        >>> f2 = u.get_fasta("P18412")
        >>> m = MUSCLE(verbose=False)
        >>> jobid = m.run(frmt="fasta", sequence=f1+f2, email="name@provider")
        >>> m.getResult(jobid, "out")

    """


    def __init__(self, verbose=False):
        url = "http://www.ebi.ac.uk/Tools/services/rest/muscle"
        self.services = REST(name='MUSCLE', url=url, verbose=verbose)
        self._parameters = None
        self._parametersDetails = {}
        self._headers = {
                "User-Agent": self.services.getUserAgent(), 
                "accept": "application/json"}

    def get_parameters(self):
        """List parameter names.

         :returns: An XML document containing a list of parameter names.

         ::

             >>> from bioservices import muscle
             >>> n = muscle.Muscle()
             >>> res = n.get_parameters()
             >>> [x.text for x in res.findAll("id")]

         .. seealso:: :attr:`parameters` to get a list of the parameters without
            need to process the XML output.
        """

        res = self.services.http_get("parameters", frmt="json", 
                headers=self._headers)
        return res['parameters']

    def _get_parameters(self):
        if self._parameters:
            return self._parameters
        else:
            # on 2 lines in case it fails, self._parameters remaisn None
            res = self.get_parameters()
            self._parameters = res
        return self._parameters
    parameters = property(_get_parameters)

    def get_parameter_details(self, parameterId):
        """Get detailed information about a parameter.

          :returns: An XML document providing details about the parameter or a list
              of values that can take the parameters if the XML could be parsed.

          For example::

              >>> n.get_parameter_details("format")

        """
        if parameterId not in self.parameters:
            raise ValueError("Invalid parameterId provided(%s). See parameters attribute" % parameterId)

        if parameterId not in self._parametersDetails.keys():
            request = "parameterdetails/" + parameterId
            res = self.services.http_get(request, frmt="json",
                headers=self._headers)
            self._parametersDetails[parameterId] = res
        return res

    def run(self, frmt=None, sequence=None, tree="none", email=None):
        """ Submit a job with the specified parameters.

        .. python ncbiblast_urllib2.py -D ENSEMBL --email "test@yahoo.com" --sequence
        .. MDSTNVRSGMKSRKKKPKTTVIDDDDDCMTCSACQSKLVKISDITKVSLDYINTMRGNTLACAACGSSLKLLNDFAS
        .. --program blastp --database uniprotkb


        .. rubric:: Compulsary arguments

        :param str frmt: input format (e.g., fasta)
        :param str sequence: query sequence. The use of fasta formatted sequence is recommended.
        :param str tree: tree type ('none','tree1','tree2')
        :param str email: a valid email address. Will be checked by the service itself.

        :return: A jobid that can be analysed with :meth:`getResult`,
            :meth:`getStatus`, ...

        The up to data values accepted for each of these parameters can be
        retrieved from the :meth:`get_parameter_details`.

        For instance,::

            from bioservices import MUSCLE
            m = MUSCLE()
            m.parameterDetails("tree")

        Example::

            jobid = m.run(frmt="fasta",
                 sequence=sequence_example,
                 email="test@yahoo.fr")

        frmt can be a list of formats::

            frmt=['fasta','clw','clwstrict','html','msf','phyi','phys']

        The returned object is a jobid, which status can be checked. It must be
        finished before analysing/geeting the results.

        .. seealso:: :meth:`getResult`

        """
        # There are compulsary arguments:
        if frmt is None or sequence is None  or email is None:
            raise ValueError("frmt, sequence and email must be provided")

        # Here, we will check the arguments values (not the type)
        # Arguments will be checked by the service itself but if we can
        # catch some before, it is better

        # FIXME: return parameters from server are not valid
        self.services.devtools.check_param_in_list(frmt,
                ['fasta', 'clw', 'clwstrict', 'html', 'msf', 'phyi', 'phys'])
        self.services.devtools.check_param_in_list(tree, ['none', 'tree1', 'tree2'])

        # parameter structure
        params = {
            'format': frmt,
            'sequence': sequence,
            'email': email}

        # headers is muscle is not required. If provided
        # by the default values from bioservices, it does not
        # work.
        headers = {}

        # IMPORTANT: use data parameter, not params !!!
        res = self.services.http_post("run", data=params,
            headers={
                "User-Agent": self.services.getUserAgent(), 
                "accept": "text/plain"})
        return res

    def get_status(self, jobid):
        """Get status of a submitted job

        :param str jobid:
        :param str jobid: a job identifier returned by :meth:`run`.
        :return: A string giving the jobid status (e.g. FINISHED).

         The values for the status are:

         *   RUNNING: the job is currently being processed.
         *   FINISHED: job has finished, and the results can then be retrieved.
         *   ERROR: an error occurred attempting to get the job status.
         *   FAILURE: the job failed.
         *   NOT_FOUND: the job cannot be found.


        """
        res = self.services.http_get("status/{}".format(jobid), frmt="txt", 
            headers={
                "User-Agent": self.services.getUserAgent(), 
                "accept": "text/plain"})
        return res

    def get_result_types(self, jobid):
        """ Get available result types for a finished job.

        :param str jobid: a job identifier returned by :meth:`run`.
        :param bool verbose: print the identifiers together with their label,
            mediaTypes, description and filesuffix.

        :return: A dictionary, which keys correspond to the identifiers. Each
            identifier is itself a dictionary containing the label, description,
            file suffix and mediaType of the identifier.
        """
        if self.get_status(jobid) != 'FINISHED':
            self.logging.warning("waiting for the job to be finished. May take a while")
            self.wait(jobid, verbose=False)
        url = 'resulttypes/' + jobid
        res = self.services.http_get(url, frmt="json", 
            headers={
                "User-Agent": self.services.getUserAgent(), 
                "accept": "application/json"})
        return [x["identifier"] for x in res['types']]

    def get_result(self, jobid, result_type):
        """ Get the job result of the specified type.


        :param str jobid: a job identifier returned by :meth:`run`.
        :param str  resultType: type of result to retrieve. See :meth:`getResultTypes`.
 
        """
        if self.get_status(jobid) != 'FINISHED': #pragma: no cover
            self.services.logging.warning("waiting for the job to be finished. May take a while")
            self.wait(jobid, verbose=False)

        if self.get_status(jobid) != "FINISHED":  #pragma: no cover
            raise ValueError("job is not finished")

        assert result_type in self.get_result_types(jobid)
        url = '/result/' + jobid + '/' + result_type


        if result_type in ['out', 'sequence', "aln-fasta", "pim", "phylotree"]:
            frmt = "txt"
        res = self.services.http_get(url, frmt=frmt,
            headers={
                "User-Agent": self.services.getUserAgent(),
                "accept": "application/json"})

        return res

    def wait(self, jobId, checkInterval=5, verbose=True):
        """This function checks the status of a jobid while it is running

        :param str jobid: a job identifier returned by :meth:`run`.
        :param int checkInterval: interval between requests in seconds.

        """
        if checkInterval < 1:  #prgma: no cover
            raise ValueError("checkInterval must be positive and less than minute")
        result = 'PENDING'
        while result == 'RUNNING' or result == 'PENDING':
            result = self.get_status(jobId)
            if verbose:
                print("WARNING: ", jobId, " is ", result, file=sys.stderr)

            if result == 'RUNNING' or result == 'PENDING':
                time.sleep(checkInterval)
        return result



#
#  This file is part of bioservices software
#
#  Copyright (c) 2013-2014 - EBI-EMBL
#
#  File author(s): 
#      
#
#  Distributed under the GPLv3 License.
#  See accompanying file LICENSE.txt or copy at
#      http://www.gnu.org/licenses/gpl-3.0.html
#
#  website: https://github.com/cokelaer/bioservices
#  documentation: http://packages.python.org/bioservices
#
##############################################################################
"""This module provides a class :class:`~BioGrid`.

.. topic:: What is BioGrid ?

    :URL: http://thebiogrid.org/
    :Service: Via the PSICQUIC class

    .. highlights::

        BioGRID is an online interaction repository with data compiled through
        comprehensive curation efforts. Our current index is version 3.2.97 and searches
        37,954 publications for 638,453 raw protein and genetic interactions from major
        model organism species. All interaction data are freely provided through our
        search index and available via download in a wide variety of standardized
        formats.

        -- From BioGrid website, Feb. 2013

"""

from bioservices import PSICQUIC
import re
from bioservices import logger
logger.name = __name__


__all__ = ["BioGRID"]


class Search(PSICQUIC):
    """ Class that carries out the actual search via psicquic.


    .. todo:: to be removed"""

    def __init__(self, data):
        super(Search, self).__init__(verbose="ERROR")
        self.data = data
        if "biogrid" in self.activeDBs:
            self.output = self.query("biogrid",self.data)
        else:
            self.logging.warning("BioGrid is not active")
            self.output = []
        self.interactors = self.get_interactors()

    def get_interactors(self):
        out = []
        for element in self.output:
            x = (re.sub(".*:","", element[2:4][0]), re.sub(".*:","", element[2:4][1]))
            out.append(tuple(sorted(x)))
        return list(set(out))


class BioGRID(object):
    """ Interface to BioGRID.

    .. doctest::

      >>> from bioservices import BioGRID
      >>> b = BioGRID(query=["map2k4","akt1"],taxId = "9606")
      >>> interactors = b.biogrid.interactors

    Examples::

        >>> from bioservices import BioGRID
        >>> b = BioGRID(query=["mtor","akt1"],taxId="9606",exP="two hybrid")
        >>> b.biogrid.interactors

    One can also query an entire organism, by using the taxid as the query::
    
        >>> b = BioGRID(query="6239")

    """

    def __init__(self, query=None, taxId=None, exP=None):

        self.query = query
        self.taxId = taxId
        self.exP = exP
        self.searchString = self._biogridSearch()
        self.biogrid = Search(self.searchString)

    def _biogridSearch(self, query=None, taxid = None, exp = None):
        """Creates a search string for the biogrid database.
    
        :param str query: the gene name(s). Can be a string or a list of strings.
        :param str taxid: the taxid for the organism of relevance. If None, 
            all organisms are choosen.
        :param str exp: the experimental protocol used to identify the interactions.
        :return: a search string for biogrid. """ 


        if query is None:
            query = self.query
        if taxid is None:
            taxid = self.taxId
        if exp is None:
            exp = self.exP

        asepNone = "%20AND%20None"
        if exp is not None:
            exp = exp.replace(" ",asepNone[:-4])
        if isinstance(query,str):
            conStr = "%s%s" % (query,asepNone)
        else:
            conStr = "%s" % ("%20OR%20".join(query))
        if taxid is not None or exp is not None:
            if isinstance(query,list):
                conStr = "(%s)" % conStr
        conStr = "%s%s%s%s%s" % (conStr,asepNone[:-4],taxid,asepNone[:-4],exp)
        conStr = conStr.replace(asepNone,"")  
        return conStr

